from ..core.expressions import MathExpression
from ..core.tree import STOP
from ..util import is_debug_mode


class BaseRule:
    """Basic rule class that visits a tree with a specified visit order."""

    @property
    def name(self):
        """Readable rule name used for debug rendering and description outputs"""
        return "Abstract Base Rule"

    @property
    def code(self) -> str:
        """Short code for debug rendering. Should be two letters."""
        return "XX"

    def find_node(self, expression: MathExpression):
        """Find the first node that can have this rule applied to it."""
        result = None

        def visit_fn(node, depth, data):
            nonlocal result
            if self.can_apply_to(node):
                result = node

            if result is not None:
                return STOP

        expression.visit_inorder(visit_fn)
        return result

    def find_nodes(self, expression: MathExpression):
        """Find all nodes in an expression that can have this rule applied to them.
        Each node is marked with it's token index in the expression, according to
        the visit strategy, and stored as `node.r_index` starting with index 0
        """
        nodes = []
        index = 0

        def visit_fn(node, depth, data):
            nonlocal nodes, index
            add = None
            node.r_index = index
            if self.can_apply_to(node):
                add = node

            index += 1
            if add:
                return nodes.append(add)

        expression.visit_inorder(visit_fn)
        return nodes

    def can_apply_to(self, node):
        """User-specified function that returns True/False if a rule can be
        applied to a given node.

        !!!warning "Performance Point"

            `can_apply_to` is called very frequently during normal operation
            and should be implemented as efficiently as possible.
        """
        return False

    def apply_to(self, node: MathExpression) -> "ExpressionChangeRule":
        """Apply the rule transformation to the given node, and return a
        ExpressionChangeRule object that captures the input/output states
        for the change."""
        # Only double-check canApply in debug mode for performance reasons
        if is_debug_mode() and not self.can_apply_to(node):
            print("Bad Apply: {}".format(node))
            print("     Root: {}".format(node.get_root()))
            raise Exception("Cannot apply {} to {}".format(self.name, node))

        return ExpressionChangeRule(self, node)


class ExpressionChangeRule:
    """Object describing the change to an expression tree from a rule transformation"""

    rule: BaseRule
    node: MathExpression
    result: MathExpression

    def __init__(self, rule, node=None):
        self.rule = rule
        self.node = node
        self._save_parent = None

    def save_parent(self, parent=None, side=None):
        """Note the parent of the node being modified, and set it as the parent of the
        rule output automatically."""
        if self.node and parent is None:
            parent = self.node.parent

        self._save_parent = parent
        if self._save_parent:
            self._save_side = side or parent.get_side(self.node)

        return self

    def done(self, node):
        """Set the result of a change to the given node. Restore the parent
        if `save_parent` was called."""
        if self._save_parent:
            self._save_parent.set_side(node, self._save_side)
        self.result = node
        return self



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tensorflow ops for common discrete-action value learning rules.

These ops define action value learning rules for discrete, scalar, action
spaces. Actions must be represented as indices in the range `[0, K)` where `K`
is the number of distinct actions.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
import tensorflow as tf
from . import base_ops
from . import indexing_ops
from . import sequence_ops

QExtra = collections.namedtuple("qlearning_extra", ["target", "td_error"])
DoubleQExtra = collections.namedtuple(
    "double_qlearning_extra", ["target", "td_error", "best_action"]
)


def qlearning(q_tm1, a_tm1, r_t, pcont_t, q_t, name="QLearning"):
    """Implements the Q-learning loss as a TensorFlow op.

  The loss is `0.5` times the squared difference between `q_tm1[a_tm1]` and
  the target `r_t + pcont_t * max q_t`.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node65.html).

  Args:
    q_tm1: Tensor holding Q-values for first timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    q_t: Tensor holding Q-values for second timestep in a batch of
      transitions, shape `[B x num_actions]`.
    name: name to prefix ops created within this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[q_tm1, q_t], [a_tm1, r_t, pcont_t]], [2, 1], name)

    # Q-learning op.
    with tf.compat.v1.name_scope(name, values=[q_tm1, a_tm1, r_t, pcont_t, q_t]):

        # Build target and select head to update.
        with tf.compat.v1.name_scope("target"):
            target = tf.stop_gradient(
                r_t + pcont_t * tf.reduce_max(input_tensor=q_t, axis=1)
            )
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - qa_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, QExtra(target, td_error))


def double_qlearning(
    q_tm1, a_tm1, r_t, pcont_t, q_t_value, q_t_selector, name="DoubleQLearning"
):
    """Implements the double Q-learning loss as a TensorFlow op.

  The loss is `0.5` times the squared difference between `q_tm1[a_tm1]` and
  the target `r_t + pcont_t * q_t_value[argmax q_t_selector]`.

  See "Double Q-learning" by van Hasselt.
  (https://papers.nips.cc/paper/3964-double-q-learning.pdf).

  Args:
    q_tm1: Tensor holding Q-values for first timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    q_t_value: Tensor of Q-values for second timestep in a batch of transitions,
      used to estimate the value of the best action, shape `[B x num_actions]`.
    q_t_selector: Tensor of Q-values for second timestep in a batch of
      transitions used to estimate the best action, shape `[B x num_actions]`.
    name: name to prefix ops created within this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[B]`
        * `td_error`: batch of temporal difference errors, shape `[B]`
        * `best_action`: batch of greedy actions wrt `q_t_selector`, shape `[B]`
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert(
        [[q_tm1, q_t_value, q_t_selector], [a_tm1, r_t, pcont_t]], [2, 1], name
    )

    # double Q-learning op.
    with tf.compat.v1.name_scope(
        name, values=[q_tm1, a_tm1, r_t, pcont_t, q_t_value, q_t_selector]
    ):

        # Build target and select head to update.
        best_action = tf.argmax(input=q_t_selector, axis=1, output_type=tf.int32)
        double_q_bootstrapped = indexing_ops.batched_index(q_t_value, best_action)
        target = tf.stop_gradient(r_t + pcont_t * double_q_bootstrapped)
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - qa_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, DoubleQExtra(target, td_error, best_action))


def persistent_qlearning(
    q_tm1, a_tm1, r_t, pcont_t, q_t, action_gap_scale=0.5, name="PersistentQLearning"
):
    """Implements the persistent Q-learning loss as a TensorFlow op.

  The loss is `0.5` times the squared difference between `q_tm1[a_tm1]` and
  `r_t + pcont_t * [(1-action_gap_scale) max q_t + action_gap_scale qa_t]`

  See "Increasing the Action Gap: New Operators for Reinforcement Learning"
  by Bellemare, Ostrovski, Guez et al. (https://arxiv.org/abs/1512.04860).

  Args:
    q_tm1: Tensor holding Q-values for first timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    q_t: Tensor holding Q-values for second timestep in a batch of
      transitions, shape `[B x num_actions]`.
      These values are used for estimating the value of the best action. In
      DQN they come from the target network.
    action_gap_scale: coefficient in [0, 1] for scaling the action gap term.
    name: name to prefix ops created within this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[q_tm1, q_t], [a_tm1, r_t, pcont_t]], [2, 1], name)
    base_ops.assert_arg_bounded(action_gap_scale, 0, 1, name, "action_gap_scale")

    # persistent Q-learning op.
    with tf.compat.v1.name_scope(name, values=[q_tm1, a_tm1, r_t, pcont_t, q_t]):

        # Build target and select head to update.
        with tf.compat.v1.name_scope("target"):
            max_q_t = tf.reduce_max(input_tensor=q_t, axis=1)
            qa_t = indexing_ops.batched_index(q_t, a_tm1)
            corrected_q_t = (1 - action_gap_scale) * max_q_t + action_gap_scale * qa_t
            target = tf.stop_gradient(r_t + pcont_t * corrected_q_t)
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - qa_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, QExtra(target, td_error))


def sarsa(q_tm1, a_tm1, r_t, pcont_t, q_t, a_t, name="Sarsa"):
    """Implements the SARSA loss as a TensorFlow op.

  The loss is `0.5` times the squared difference between `q_tm1[a_tm1]` and
  the target `r_t + pcont_t * q_t[a_t]`.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node64.html.)

  Args:
    q_tm1: Tensor holding Q-values for first timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    q_t: Tensor holding Q-values for second timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_t: Tensor holding action indices for second timestep, shape `[B]`.
    name: name to prefix ops created within this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[q_tm1, q_t], [a_t, r_t, pcont_t]], [2, 1], name)

    # SARSA op.
    with tf.compat.v1.name_scope(name, values=[q_tm1, a_tm1, r_t, pcont_t, q_t, a_t]):

        # Select head to update and build target.
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)
        qa_t = indexing_ops.batched_index(q_t, a_t)
        target = tf.stop_gradient(r_t + pcont_t * qa_t)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - qa_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, QExtra(target, td_error))


def sarse(q_tm1, a_tm1, r_t, pcont_t, q_t, probs_a_t, debug=False, name="Sarse"):
    """Implements the SARSE (Expected SARSA) loss as a TensorFlow op.

  The loss is `0.5` times the squared difference between `q_tm1[a_tm1]` and
  the target `r_t + pcont_t * (sum_a probs_a_t[a] * q_t[a])`.

  See "A Theoretical and Empirical Analysis of Expected Sarsa" by Seijen,
  van Hasselt, Whiteson et al.
  (http://www.cs.ox.ac.uk/people/shimon.whiteson/pubs/vanseijenadprl09.pdf).

  Args:
    q_tm1: Tensor holding Q-values for first timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    q_t: Tensor holding Q-values for second timestep in a batch of
      transitions, shape `[B x num_actions]`.
    probs_a_t: Tensor holding action probabilities for second timestep,
      shape `[B x num_actions]`.
    debug: Boolean flag, when set to True adds ops to check whether probs_a_t
      is a batch of (approximately) valid probability distributions.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert(
        [[q_tm1, q_t, probs_a_t], [a_tm1, r_t, pcont_t]], [2, 1], name
    )

    # SARSE (Expected SARSA) op.
    with tf.compat.v1.name_scope(
        name, values=[q_tm1, a_tm1, r_t, pcont_t, q_t, probs_a_t]
    ):

        # Debug ops.
        deps = []
        if debug:
            cumulative_prob = tf.reduce_sum(input_tensor=probs_a_t, axis=1)
            almost_prob = tf.less(tf.abs(tf.subtract(cumulative_prob, 1.0)), 1e-6)
            deps.append(
                tf.Assert(
                    tf.reduce_all(input_tensor=almost_prob),
                    ["probs_a_t tensor does not sum to 1", probs_a_t],
                )
            )

        # With dependency on possible debug ops.
        with tf.control_dependencies(deps):

            # Select head to update and build target.
            qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)
            target = tf.stop_gradient(
                r_t
                + pcont_t
                * tf.reduce_sum(input_tensor=tf.multiply(q_t, probs_a_t), axis=1)
            )

            # Temporal difference error and loss.
            # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
            td_error = target - qa_tm1
            loss = 0.5 * tf.square(td_error)
            return base_ops.LossOutput(loss, QExtra(target, td_error))


def qlambda(q_tm1, a_tm1, r_t, pcont_t, q_t, lambda_, name="GeneralizedQLambda"):
    """Implements Peng's and Watkins' Q(lambda) loss as a TensorFlow op.

  This function is general enough to implement both Peng's and Watkins'
  Q-lambda algorithms.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node78.html).

  Args:
    q_tm1: `Tensor` holding a sequence of Q-values starting at the first
      timestep; shape `[T, B, num_actions]`
    a_tm1: `Tensor` holding a sequence of action indices, shape `[T, B]`
    r_t: Tensor holding a sequence of rewards, shape `[T, B]`
    pcont_t: `Tensor` holding a sequence of pcontinue values, shape `[T, B]`
    q_t: `Tensor` holding a sequence of Q-values for second timestep;
      shape `[T, B, num_actions]`. In a target network setting,
      this quantity is often supplied by the target network.
    lambda_: a scalar or `Tensor` of shape `[T, B]`
      specifying the ratio of mixing between bootstrapped and MC returns;
      if lambda_ is the same for all time steps then the function implements
      Peng's Q-learning algorithm; if lambda_ = 0 at every sub-optimal action
      and a constant otherwise, then the function implements Watkins'
      Q-learning algorithm. Generally lambda_ can be a Tensor of any values
      in the range [0, 1] supplied by the user.
    name: a name of the op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[T, B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[T, B]`.
        * `td_error`: batch of temporal difference errors, shape `[T, B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[q_tm1, q_t]], [3], name)
    if isinstance(lambda_, tf.Tensor) and lambda_.get_shape().ndims > 0:
        base_ops.wrap_rank_shape_assert([[a_tm1, r_t, pcont_t, lambda_]], [2], name)
    else:
        base_ops.wrap_rank_shape_assert([[a_tm1, r_t, pcont_t]], [2], name)

    # QLambda op.
    with tf.compat.v1.name_scope(name, values=[q_tm1, a_tm1, r_t, pcont_t, q_t]):

        # Build target and select head to update.
        with tf.compat.v1.name_scope("target"):
            state_values = tf.reduce_max(input_tensor=q_t, axis=2)
            target = sequence_ops.multistep_forward_view(
                r_t, pcont_t, state_values, lambda_, back_prop=False
            )
            target = tf.stop_gradient(target)
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - qa_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, QExtra(target, td_error))


def qv_learning(q_tm1, a_tm1, r_t, pcont_t, v_t, name="QVLearning"):
    """Implements the QV loss as a TensorFlow op.

  The loss is `0.5` times the squared difference between `q_tm1[a_tm1]` and
  the target `r_t + pcont_t * v_t`, where `v_t` is separately learned through
  temporal difference learning (c.f. `value_ops.td_learning`).

  See "Two Novel On-policy Reinforcement Learning Algorithms based on
  TD(lambda)-methods" by Wiering and van Hasselt
  (https://ieeexplore.ieee.org/abstract/document/4220845.)

  Args:
    q_tm1: Tensor holding Q-values for first timestep in a batch of
      transitions, shape `[B x num_actions]`.
    a_tm1: Tensor holding action indices, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    v_t: Tensor holding state-values for second timestep in a batch of
      transitions, shape `[B]`.
    name: name to prefix ops created within this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `q_tm1[a_tm1]`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[q_tm1], [a_tm1, r_t, pcont_t, v_t]], [2, 1], name)

    # QV op.
    with tf.compat.v1.name_scope(name, values=[q_tm1, a_tm1, r_t, pcont_t, v_t]):

        # Build target and select head to update.
        with tf.compat.v1.name_scope("target"):
            target = tf.stop_gradient(r_t + pcont_t * v_t)
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - qa_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, QExtra(target, td_error))




# Based on spaCy's model loading utilities and CLI "package" command:
#  - https://github.com/explosion/spaCy/blob/master/spacy/util.py
#  - https://github.com/explosion/spaCy/blob/master/spacy/cli/package.py
"""Mathy implements a model loading and packaging scheme [based on the one from 
**spaCy**](https://spacy.io/){target=\_blank}. 

It can load models from **pypi packages** as well as **local folders** and **system links**.

!!! tip "spaCy is great for NLP tasks"

    Machine learning can often feel complex and hard to approach. spaCy is an open-source library
    for **Natural Language Processing** that is **ready for production**, **easy-to-use**, and **blazing fast**.

    With spaCy it's simple to get started working with text inputs:

    ```python
    import spacy

    nlp = spacy.load("en_core_web_sm")
    doc = nlp("justin and alex are both terrible at math")
    people = [t for t in doc if t.pos_ == "PROPN"]
    print(f"People found: {people}")
    ```

    Outputs:
    ```
    People found: [justin, alex]
    ```

    By looking at the **Parts of Speech** (referenced as `t.pos_`) we can find the people
    referenced in input.

    spaCy does much more than this, with examples that run inside the website. 
    Check it out: **[https://spacy.io](https://spacy.io){target=\_blank}**

"""
import importlib
import shutil
from pathlib import Path
from typing import Any, Optional, Union

import srsly
from wasabi import get_raw_input, msg

from . import about
from .api import Mathy

_data_path = Path(__file__).parent / "data"


REQUIRED_META_KEYS = ["units", "embedding_units", "lstm_units", "version"]
REQUIRED_MODEL_FILES = ["model.h5", "model.optimizer", "model.config.json"]


def load_model(name: str, **overrides) -> Mathy:
    """Load a model from a shortcut link, package or data path.

    # Arguments
    name (str): Package name, shortcut link or model path.
    overrides (kwargs): Specific overrides, like how many MCTS sims to use

    # Raises
    ValueError: If *name* is not recognized loadable package or model folder.

    # Returns
    (Mathy): Mathy class with the loaded model.
    """
    if isinstance(name, str):
        if is_package(name):  # installed as package
            return load_model_from_package(name, **overrides)
        if Path(name).exists():  # path to model data directory
            return load_model_from_path(Path(name), **overrides)
    elif hasattr(name, "exists"):  # Path or Path-like to model data
        return load_model_from_path(name, **overrides)
    raise ValueError(f"Unrecognized model input: {name}")


def load_model_from_package(name, **overrides):
    """Load a model from an installed package."""
    cls = importlib.import_module(name)
    return cls.load(**overrides)


def load_model_from_path(model_path: Path, meta: dict = None, **overrides) -> Mathy:
    """Load a model from a data directory path.
    
    # Arguments
    model_path (Path): The model folder to load from.

    # Returns
    (Mathy): a Mathy instance.
    """
    if not meta:
        meta = get_model_meta(model_path)
    mt = Mathy(model_path=str(model_path), **overrides)
    return mt


def load_model_from_init_py(init_file: Union[Path, str], **overrides):
    """Helper function to use in the `load()` method of a model package's
    __init__.py.

    # Arguments
    init_file (unicode): Path to model's __init__.py, i.e. `__file__`.
    **overrides: Specific overrides, like pipeline components to disable.

    # Raises
    ValueError: If the model path does not exist.
    
    # Returns
    (Language): `Language` class with loaded model.
    """
    model_path = Path(init_file).parent
    if not model_path.exists():
        raise ValueError(f"model path does not exist: {model_path}")
    meta = get_model_meta(model_path)
    data_path = model_path
    return load_model_from_path(data_path, meta, **overrides)


def get_model_meta(model_path: Path):
    """Get model meta.json from a directory path and validate its contents.
    path (unicode or Path): Path to model directory.

    # Raises
    ValueError: If **model_path** does not point to a valid folder.
    ValueError: If **model_path** does not have a `model.config.json` file.
    ValueError: If any required settings are missing from the meta file.

    # Returns
    (dict): The model's meta data.
    """
    model_path = Path(model_path)
    if not model_path.exists():
        raise ValueError(f"cannot get meta from invalid model path: {model_path}")
    meta_path = model_path / "model.config.json"
    if not meta_path.is_file():
        raise ValueError(f"invalid meta file: {meta_path}")
    meta = srsly.read_json(meta_path)
    for setting in REQUIRED_META_KEYS:
        if setting not in meta or not meta[setting]:
            raise ValueError(f"meta file missing required setting: {setting}")
    return meta


def is_package(name: str) -> bool:
    """Check if string maps to a package installed via pip.

    # Arguments
    name (str): Name of package.

    # Returns
    (bool): True if installed package, False if not.
    """
    import pkg_resources

    name = name.lower()  # compare package name against lowercase name
    packages = pkg_resources.working_set.by_key.keys()  # type:ignore
    for package in packages:
        if package.lower().replace("-", "_") == name:
            return True
    return False


def package(
    model_name: str,
    input_dir: Union[str, Path],
    output_dir: Union[str, Path],
    meta_path: Union[str, Path] = None,
    force: bool = False,
) -> str:
    """Generate a Python package from a Mathy model.
    
    A new directory will be created in the specified output directory, and model data will be copied over.

    # Arguments
    model_name: the lower-case with underscores name for the model, e.g. "mathy_alpha_sm".
    input_dir: The folder containing a Mathy model to use as an input.
    output_dir: The folder to put the packaged model in.
    force: Delete any existing model in the output_dir rather than throw an error when this is true.

    # Raises
    SystemExit: If the input path does not exist.
    SystemExit: If the input path has no model.config.json file.
    SystemExit: If the model.config.json file is missing required keys.
    SystemExit: If output folder exists and `force` is False.

    # Returns
    (str): The subfolder of the output path that contains the pypi package source.
    """
    input_path = Path(input_dir)
    output_path = Path(output_dir)
    if meta_path is not None:
        meta_path = Path(meta_path)
    if not input_path or not input_path.exists():
        msg.fail("Can't locate model data", input_path, exits=1)
    if meta_path and not meta_path.exists():
        msg.fail("Can't find model model.config.json", meta_path, exits=1)
    meta_path = meta_path or input_path / "model.config.json"
    if meta_path.is_file():
        meta = srsly.read_json(meta_path)
        msg.good("Loaded model.config.json from file", meta_path)
    meta["mathy_version"] = f">={about.__version__},<1.0.0"
    meta["name"] = model_name
    for key in REQUIRED_META_KEYS:
        if key not in meta or meta[key] == "":
            msg.fail(
                "No '{}' setting found in model.config.json".format(key),
                "This setting is required to build your package.",
                exits=1,
            )
    main_path = output_path / model_name
    package_path = main_path
    if package_path.exists():
        if force:
            shutil.rmtree(str(package_path))
        else:
            msg.fail(
                title="Package directory already exists",
                text="Please delete the directory and try again, or use the "
                "`--force` flag to overwrite existing directories.",
                exits=1,
            )
    Path.mkdir(package_path, parents=True, exist_ok=True)
    for f in REQUIRED_MODEL_FILES:
        file_name: Path = input_path / f
        if not file_name.exists():
            msg.fail(
                f"Input path '{input_path}' is missing a required file: '{f}'",
                "This file is required to build your package.",
                exits=1,
            )
        shutil.copyfile(file_name, main_path / f)
    create_file(output_path / "model.config.json", srsly.json_dumps(meta, indent=2))
    create_file(output_path / "setup.py", TEMPLATE_SETUP)
    create_file(package_path / "__init__.py", TEMPLATE_INIT)
    msg.good("Successfully created package '{}'".format(package_path), main_path)
    msg.text("To build the package, run `python setup.py sdist` in this directory.")
    return str(package_path)


def create_file(file_path: Path, contents: str):
    file_path.touch()
    file_path.open("w", encoding="utf-8").write(contents)


TEMPLATE_SETUP = """
#!/usr/bin/env python
# coding: utf8
from typing import List
import io
import json
from os import path, walk
from shutil import copy
from setuptools import setup


def load_meta(fp):
    with io.open(fp, encoding="utf8") as f:
        return json.load(f)


def list_files(data_dir: str, root: str) -> List[str]:
    output = []
    package_dir = path.join(root, data_dir)
    for folder, _, filenames in walk(package_dir):
        if "__pycache__" in folder:
            continue
        for filename in filenames:
            if not filename.startswith("."):
                output.append(path.join(folder, filename))
    rel_output = [path.relpath(p, package_dir) for p in output]
    return rel_output


def list_requirements(meta):
    parent_package = meta.get("parent_package", "mathy")
    requirements = [parent_package + meta["mathy_version"]]
    if "setup_requires" in meta:
        requirements += meta["setup_requires"]
    if "requirements" in meta:
        requirements += meta["requirements"]
    return requirements


def setup_package():
    root = path.abspath(path.dirname(__file__))
    meta_path = path.join(root, "model.config.json")
    meta = load_meta(meta_path)
    model_name = meta["name"]
    model_dir = path.join(model_name)
    setup(
        name=model_name,
        description=meta["description"],
        author=meta["author"],
        author_email=meta["email"],
        url=meta["url"],
        version=meta["version"],
        license=meta["license"],
        packages=[model_name],
        package_data={model_name: list_files(model_dir, root)},
        install_requires=list_requirements(meta),
        zip_safe=False,
    )


if __name__ == "__main__":
    setup_package()""".strip()


TEMPLATE_INIT = """
# coding: utf8
from __future__ import unicode_literals
from pathlib import Path
from mathy.models import load_model_from_init_py, get_model_meta
__version__ = get_model_meta(Path(__file__).parent)['version']
def load(**overrides):
    return load_model_from_init_py(__file__, **overrides)
""".strip()



from ..mathy.core.tokenizer import Token
from ..mathy.core.tree import BinaryTreeNode
from ..mathy.core.expressions import (
    ConstantExpression,
    VariableExpression,
    AddExpression,
)
from ..mathy.core.parser import ExpressionParser


def test_parser_to_string():
    parser = ExpressionParser()
    expects = [
        {
            "input": "(-2.257893300159429e+16h^2 * v) * j^4",
            "output": "(-2.257893300159429e + 16h^2 * v) * j^4",
        },
        {"input": "1f + 98i + 3f + 14t", "output": "1f + 98i + 3f + 14t"},
        {"input": "4x * p^(1 + 3) * 12x^2", "output": "4x * p^(1 + 3) * 12x^2"},
        {"input": "(5 * 3) * (32 / 7)", "output": "(5 * 3) * 32 / 7"},
        {"input": "7 - 5 * 3 * (2^7)", "output": "7 - 5 * 3 * 2^7"},
        {"input": "(8x^2 * 9b) * 7", "output": "(8x^2 * 9b) * 7"},
        {"input": "(8 * 9b) * 7", "output": "(8 * 9b) * 7"},
        {"input": "7 - (5 * 3) * (32 / 7)", "output": "7 - (5 * 3) * 32 / 7"},
        {"input": "7 - (5 - 3) * (32 - 7)", "output": "7 - (5 - 3) * (32 - 7)"},
        {"input": "(7 - (5 * 3)) * (32 - 7)", "output": "(7 - 5 * 3) * (32 - 7)"},
    ]
    # Test to make sure parens are preserved in output when they are meaningful
    for expect in expects:
        expression = parser.parse(expect["input"])
        out_str = str(expression)
        assert out_str == expect["output"]


def test_mult_exp_precedence():
    """should respect order of operations with factor parsing"""
    parser = ExpressionParser()
    expression = parser.parse("4x^2")
    val = expression.evaluate({"x": 2})
    # 4x^2 should evaluate to 16 with x=2
    assert val == 16

    expression = parser.parse("7 * 10 * 6x * 3x + 5x")
    assert expression is not None



__title__ = "mathy"
__version__ = "0.6.5"
__summary__ = "Mathy - RL environments for solving math problems step-by-step"
__uri__ = "https://mathy.ai"
__author__ = "Justin DuJardin"
__email__ = "justin@dujardinconsulting.com"
__license__ = "All rights reserved"



from typing import List, Optional

import tensorflow as tf


class DenseNetBlock(tf.keras.layers.Layer):
    """DenseNet like block layer that concatenates inputs with extracted features
    and feeds them into all future tower layers."""

    def __init__(
        self,
        units=128,
        num_layers=2,
        use_shared=False,
        activation="relu",
        normalization="batch",
        **kwargs,
    ):
        super(DenseNetBlock, self).__init__(**kwargs)
        self.units = units
        self.use_shared = use_shared
        self.activate = tf.keras.layers.Activation("relu", name="relu")
        if normalization == "batch":
            self.normalize = tf.keras.layers.BatchNormalization()
        elif normalization == "layer":
            self.normalize = tf.keras.layers.LayerNormalization()
        else:
            raise ValueError(f"unkknown layer normalization style: {normalization}")
        self.dense = tf.keras.layers.Dense(
            units, use_bias=False, name=f"{self.name}_dense"
        )
        self.concat = tf.keras.layers.Concatenate(name=f"{self.name}_concat")
        self.activate = tf.keras.layers.Activation(activation, name="activate")

    def do_op(self, input_tensor: tf.Tensor, previous_tensors: List[tf.Tensor]):
        inputs = previous_tensors[:]
        inputs.append(input_tensor)
        # concatenate the input and previous layer inputs
        if (len(inputs)) > 1:
            input_tensor = self.concat(inputs)
        return self.normalize(self.activate(self.dense(input_tensor)))

    def call(self, input_tensor: tf.Tensor, previous_tensors: List[tf.Tensor]):
        if self.use_shared:
            with tf.compat.v1.variable_scope(
                self.name, reuse=True, auxiliary_name_scope=False
            ):
                return self.do_op(input_tensor, previous_tensors)
        else:
            with tf.compat.v1.variable_scope(self.name):
                return self.do_op(input_tensor, previous_tensors)


class DenseNetStack(tf.keras.layers.Layer):
    """DenseNet like stack of residually connected layers where the input and all
    of the previous outputs are provided as the input to each layer:

        https://arxiv.org/pdf/1608.06993.pdf

    From the paper: "Crucially, in contrast to ResNets, we never combine features
    through summation before they are passed into a layer; instead, we combine features
    by concatenating them"
    """

    def __init__(
        self,
        units=64,
        num_layers=4,
        layer_scaling_factor=0.75,
        share_weights=False,
        activation="relu",
        output_transform: Optional[tf.keras.layers.Layer] = None,
        normalization_style: str = "layer",
        **kwargs,
    ):
        self.units = units
        self.layer_scaling_factor = layer_scaling_factor
        self.num_layers = num_layers
        self.output_transform = output_transform
        if activation is not None:
            self.activate = tf.keras.layers.Activation(activation)
        else:
            self.activate = None
        self.concat = tf.keras.layers.Concatenate()
        block_units = int(self.units)
        self.dense_stack = [
            DenseNetBlock(
                block_units,
                name="densenet_0",
                use_shared=share_weights,
                normalization=normalization_style,
            )
        ]
        for i in range(self.num_layers - 1):
            block_units = int(block_units * self.layer_scaling_factor)
            self.dense_stack.append(
                DenseNetBlock(
                    block_units,
                    name=f"densenet_{i + 1}",
                    use_shared=share_weights,
                    normalization=normalization_style,
                )
            )
        super(DenseNetStack, self).__init__(**kwargs)

    def call(self, input_tensor: tf.Tensor):
        stack_inputs: List[tf.Tensor] = []
        root_input = input_tensor
        # Iterate the stack and call each layer, also inputting a list
        # of all the previous stack layers outputs.
        for layer in self.dense_stack:
            # Save reference to input
            prev_tensor = input_tensor
            # Apply densnet block
            input_tensor = layer(input_tensor, stack_inputs)
            # Append the current input to the list of previous input tensors
            stack_inputs.append(prev_tensor)
        output = self.concat([input_tensor, root_input])
        if self.output_transform is not None:
            output = self.output_transform(output)
        return self.activate(output) if self.activate is not None else output



from ..mathy.core.expressions import (
    AddExpression,
    ConstantExpression,
    DivideExpression,
    MultiplyExpression,
    PowerExpression,
    VariableExpression,
)
from ..mathy.core.parser import ExpressionParser
from ..mathy.rules import (
    AssociativeSwapRule,
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    DistributiveMultiplyRule,
)
from ..mathy.util import (
    TermEx,
    get_sub_terms,
    get_term_ex,
    has_like_terms,
    is_preferred_term_form,
)


def test_get_sub_terms():
    expectations = [
        ("70656 * (x^2 * z^6)", 2),
        ("4x^2 * z^6 * y", 3),
        ("2x^2", 1),
        ("x^2", 1),
        ("2", 1),
    ]
    invalid_expectations = [
        # can't have more than one term
        ("4 + 4", False)
    ]
    parser = ExpressionParser()
    for text, output in expectations + invalid_expectations:
        exp = parser.parse(text)
        sub_terms = get_sub_terms(exp)
        if output is False:
            assert text == text and sub_terms == output
        else:
            assert text == text and len(sub_terms) == output


def test_get_term_ex():
    examples = [
        ("4x^2", TermEx(4, "x", 2)),
        ("4x", TermEx(4, "x", None)),
        ("x", TermEx(None, "x", None)),
        # TODO: non-natural term forms? If this is supported we can drop the other
        #       get_term impl maybe?
        # ("x * 2", TermEx(2, "x", None)),
    ]
    parser = ExpressionParser()
    for input, expected in examples:
        expr = parser.parse(input)
        assert input == input and get_term_ex(expr) == expected


def test_is_preferred_term_form():
    examples = [
        ("b * (44b^2)", False),
        ("z * (1274z^2)", False),
        ("4x * z", True),
        ("z * 4x", True),
        ("2x * x", False),
        ("29y", True),
        ("z", True),
        ("z * 10", False),
        ("4x^2", True),
    ]
    parser = ExpressionParser()
    for input, expected in examples:
        expr = parser.parse(input)
        assert input == input and is_preferred_term_form(expr) == expected


def test_has_like_terms():
    examples = [
        ("14 + 6y + 7x + x * (3y)", False),
        ("b * (44b^2)", False),
        ("z * (1274z^2)", False),
        ("100y * x + 2", False),
    ]
    parser = ExpressionParser()
    for input, expected in examples:
        expr = parser.parse(input)
        assert input == input and has_like_terms(expr) == expected



from ..core.expressions import (
    AddExpression,
    MultiplyExpression,
    ConstantExpression,
    VariableExpression,
    PowerExpression,
)
from ..core.rule import BaseRule
from ..util import unlink


class DistributiveMultiplyRule(BaseRule):
    r"""
    Distributive Property
    `a(b + c) = ab + ac`

    The distributive property can be used to expand out expressions
    to allow for simplification, as well as to factor out common properties of terms.

    **Distribute across a group**

    This handles the `a(b + c)` conversion of the distributive property, which
    distributes `a` across both `b` and `c`.

    *note: this is useful because it takes a complex Multiply expression and
    replaces it with two simpler ones.  This can expose terms that can be
    combined for further expression simplification.*

                                 +
             *                  / \
            / \                /   \
           /   \              /     \
          a     +     ->     *       *
               / \          / \     / \
              /   \        /   \   /   \
             b     c      a     b a     c
    """

    @property
    def name(self):
        return "Distributive Multiply"

    @property
    def code(self):
        return "DM"

    def can_apply_to(self, expression):
        if isinstance(expression, MultiplyExpression):
            if expression.right and isinstance(expression.left, AddExpression):
                return True

            if expression.left and isinstance(expression.right, AddExpression):
                return True

            return False

        return False

    def apply_to(self, node):
        change = super().apply_to(node).save_parent()

        if isinstance(node.left, AddExpression):
            a = node.right
            b = node.left.left
            c = node.left.right
        else:
            a = node.left
            b = node.right.left
            c = node.right.right

        a = unlink(a)

        # If the operands for either multiplication can be expressed
        # in a "natural" order, do so like a human would.
        a_exp_var = (
            isinstance(a, PowerExpression)
            and isinstance(a.right, VariableExpression)
            or isinstance(a.left, VariableExpression)
        )
        a_var: bool = a_exp_var or isinstance(a, VariableExpression)
        b_const: bool = isinstance(b, ConstantExpression)
        c_const: bool = isinstance(c, ConstantExpression)
        if a_var and b_const:
            ab = MultiplyExpression(b.clone(), a.clone())
        else:
            ab = MultiplyExpression(a.clone(), b.clone())

        if a_var and c_const:
            ac = MultiplyExpression(c.clone(), a.clone())
        else:
            ac = MultiplyExpression(a.clone(), c.clone())
        result = AddExpression(ab, ac)

        [n.set_changed() for n in [node, a, ab, ac, result]]

        change.done(result)
        return change



from typing import List, Tuple

import pytest

from ..mathy.envs.poly_simplify import PolySimplify
from ..mathy.state import MathyEnvState, rnn_placeholder_state


def test_mathy_features_from_state():
    with pytest.raises(ValueError):
        state = MathyEnvState()
    state = MathyEnvState(problem="4x+2x")
    assert state.to_observation([]) is not None


def test_mathy_features_hierarchy():
    """Verify that the observation generated encodes hierarchy properly
    so the model can determine the precise nodes to act on"""

    diff_pairs: List[Tuple[str, str]] = [
        ("4x + (3u + 7x + 3u) + 4u", "4x + 3u + 7x + 3u + 4u"),
        ("7c * 5", "7 * (c * 5)"),
        ("5v + 20b + (10v + 7b)", "5v + 20b + 10v + 7b"),
        ("5s + 60 + 12s + s^2", "5s + 60 + (12s + s^2)"),
    ]
    env = PolySimplify()
    rnn_state = rnn_placeholder_state(128)

    for one, two in diff_pairs:
        state_one = MathyEnvState(problem=one)
        obs_one = state_one.to_observation(env.get_valid_moves(state_one), rnn_state)

        state_two = MathyEnvState(problem=two)
        obs_two = state_two.to_observation(env.get_valid_moves(state_two), rnn_state)

        assert obs_one.nodes != obs_two.nodes



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Flattened namespace for ."""

from .action_value_ops import double_qlearning
from .action_value_ops import persistent_qlearning
from .action_value_ops import qlambda
from .action_value_ops import qlearning
from .action_value_ops import qv_learning
from .action_value_ops import sarsa
from .action_value_ops import sarse
from .base_ops import assert_rank_and_shape_compatibility
from .base_ops import best_effort_shape
from .clipping_ops import huber_loss
from .discrete_policy_gradient_ops import discrete_policy_entropy_loss
from .discrete_policy_gradient_ops import discrete_policy_gradient
from .discrete_policy_gradient_ops import discrete_policy_gradient_loss
from .discrete_policy_gradient_ops import sequence_advantage_actor_critic_loss
from .dist_value_ops import categorical_dist_double_qlearning
from .dist_value_ops import categorical_dist_qlearning
from .dist_value_ops import categorical_dist_td_learning
from .dpg_ops import dpg
from .indexing_ops import batched_index
from .periodic_ops import periodically
from .policy_gradient_ops import policy_entropy_loss
from .policy_gradient_ops import policy_gradient
from .policy_gradient_ops import policy_gradient_loss
from .policy_gradient_ops import sequence_a2c_loss
from .retrace_ops import retrace
from .retrace_ops import retrace_core
from .sequence_ops import multistep_forward_view
from .sequence_ops import scan_discounted_sum
from .target_update_ops import periodic_target_update
from .target_update_ops import update_target_variables
from .value_ops import generalized_lambda_returns
from .value_ops import qv_max
from .value_ops import td_lambda
from .value_ops import td_learning
from .vtrace_ops import vtrace_from_importance_weights
from .vtrace_ops import vtrace_from_logits

__version__ = "1.33.7"



from ..mathy import MathyEnvState


def test_math_state():
    state = MathyEnvState(problem="4+4")
    assert state is not None


def test_math_state_encode_player():
    env_state = MathyEnvState(problem="4x+2")
    env_state = env_state.get_out_state(
        problem="2+4x", focus_index=3, moves_remaining=10, action=0
    )
    agent = env_state.agent
    assert agent.problem == "2+4x"
    assert agent.moves_remaining == 10
    assert agent.action == 0
    assert agent.focus_index == 3


def test_math_state_to_observation():
    """to_observation has defaults to allow calling with no arguments"""
    env_state = MathyEnvState(problem="4x+2")
    assert env_state.to_observation() is not None



from ..poly_combine_in_place import PolyCombineInPlace
from ...types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register

#
# Combine like terms without commuting
#


class GymPolynomialCombineInPlace(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs):
        super(GymPolynomialCombineInPlace, self).__init__(
            env_class=PolyCombineInPlace,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class PolynomialCombineInPlaceEasy(GymPolynomialCombineInPlace):
    def __init__(self, **kwargs):
        super(PolynomialCombineInPlaceEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class PolynomialCombineInPlaceNormal(GymPolynomialCombineInPlace):
    def __init__(self, **kwargs):
        super(PolynomialCombineInPlaceNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class PolynomialCombineInPlaceHard(GymPolynomialCombineInPlace):
    def __init__(self, **kwargs):
        super(PolynomialCombineInPlaceHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(
    id="mathy-poly-combine-easy-v0",
    entry_point="mathy.envs.gym:PolynomialCombineInPlaceEasy",
)
safe_register(
    id="mathy-poly-combine-normal-v0",
    entry_point="mathy.envs.gym:PolynomialCombineInPlaceNormal",
)
safe_register(
    id="mathy-poly-combine-hard-v0",
    entry_point="mathy.envs.gym:PolynomialCombineInPlaceHard",
)



"""Tree Layout
---

In order to help visualize, understand, and debug math trees and transformations to
them, Mathy implements a
[Reingold-Tilford](https://reingold.co/tidier-drawings.pdf){target=_blank} layout
algorithm that works with expression trees. It produces beautiful trees like:

`mathy:(2x^3 + y)(14 + 2.3y)`

"""
from .tree import BinaryTreeNode


class TreeLayout:
    """Calculate a visual layout for input trees."""

    def layout(
        self,
        node: BinaryTreeNode,
        unit_x_multiplier: float = 1.0,
        unit_y_multiplier: float = 1.0,
    ) -> "TreeMeasurement":
        """Assign x/y values to all nodes in the tree, and return an object containing
        the measurements of the tree.

        Returns a TreeMeasurement object that describes the bounds of the tree"""
        self.measure(node)
        return self.transform(node, 0, unit_x_multiplier, unit_y_multiplier)

    def measure(
        self, node: BinaryTreeNode = None, level=0, extremes: "TidierExtreme" = None,
    ) -> "TreeLayout":
        if extremes is None:
            extremes = TidierExtreme()

        # left and right subtree extreme leaf nodes
        left_extremes = TidierExtreme()
        right_extremes = TidierExtreme()

        # separation at the root of the current subtree and current level.
        current_separation = 0.0
        root_separation = 0.0
        min_separation = 1.0

        # The offset from left/right children to the root of the current subtree.
        left_offset_sum = 0.0
        right_offset_sum = 0.0

        # Avoid selecting as extreme
        if not node:
            if extremes.left is not None:
                extremes.left.level = -1

            if extremes.right is not None:
                extremes.right.level = -1

            return self

        # Assign the `node.y`, note the left/right child nodes, and recurse
        node.y = level
        left = node.left
        right = node.right
        self.measure(left, level + 1, left_extremes)
        self.measure(right, level + 1, right_extremes)

        # A leaf is both the leftmost and rightmost node on the lowest level of the
        # subtree consisting of itself.
        if not node.right and not node.left:
            node.offset = 0
            extremes.right = extremes.left = node
            return self

        # if only a single child, assign the next available offset and return.
        if not node.right or not node.left:
            node.offset = min_separation
            extremes.right = extremes.left = node.left if node.left else node.right
            return self

        # Set the current separation to the minimum separation for the root of the
        # subtree.
        current_separation = min_separation
        left_offset_sum = right_offset_sum = 0

        # Traverse the subtrees until one of them is exhausted, pushing them apart
        # as needed.
        loops = 0
        while left and right:
            loops = loops + 1
            if loops > 100000:
                raise Exception("An impossibly large tree perhaps?")

            if current_separation < min_separation:
                root_separation += min_separation - current_separation
                current_separation = min_separation

            if left.right and left.offset:
                left_offset_sum += left.offset
                current_separation -= left.offset
                left = getattr(left, "thread", left.right)
            elif left.offset is not None:
                left_offset_sum -= left.offset
                current_separation += left.offset
                left = getattr(left, "thread", left.left)

            if right.left and right.offset:
                right_offset_sum -= right.offset
                current_separation -= right.offset
                right = getattr(right, "thread", right.left)
            elif right.offset is not None:
                right_offset_sum += right.offset
                current_separation += right.offset
                right = getattr(right, "thread", right.right)

        # Set the root offset, and include it in the accumulated offsets.
        node.offset = (root_separation + 1) / 2
        left_offset_sum -= node.offset
        right_offset_sum += node.offset

        # Update right and left extremes
        right_left_level = getattr(right_extremes.left, "level", -1)
        left_left_level = getattr(left_extremes.left, "level", -1)
        if right_left_level > left_left_level or not node.left:
            extremes.left = right_extremes.left
            if extremes.left:
                extremes.left.offset += node.offset

        else:
            extremes.left = left_extremes.left
            if extremes.left:
                extremes.left.offset -= node.offset

        left_right_level = getattr(left_extremes.right, "level", -1)
        right_right_level = getattr(right_extremes.right, "level", -1)
        if left_right_level > right_right_level or not node.right:
            extremes.right = left_extremes.right
            if extremes.right:
                extremes.right.offset -= node.offset

        else:
            extremes.right = right_extremes.right
            if extremes.right:
                extremes.right.offset += node.offset

        # If the subtrees have uneven heights, check to see if they need to be
        # threaded.  If threading is required, it will affect only one node.
        if left and left != node.left and right_extremes and right_extremes.right:
            right_extremes.right.thread = left
            right_extremes.right.offset = abs(
                right_extremes.right.offset + node.offset - left_offset_sum
            )
        elif right and right != node.right and left_extremes and left_extremes.left:
            left_extremes.left.thread = right
            left_extremes.left.offset = abs(
                left_extremes.left.offset - node.offset - right_offset_sum
            )

        return self

    def transform(
        self,
        node: BinaryTreeNode = None,
        x=0,
        unit_x_multiplier=1,
        unit_y_multiplier=1,
        measure=None,
    ) -> "TreeMeasurement":
        """Transform relative to absolute coordinates, and measure the bounds of the tree.

        Return a measurement of the tree in output units."""
        if measure is None:
            measure = TreeMeasurement()
        if not node:
            return measure

        node.x = x * unit_x_multiplier
        node.y *= unit_y_multiplier
        self.transform(
            node.left, x - node.offset, unit_x_multiplier, unit_y_multiplier, measure
        )
        self.transform(
            node.right, x + node.offset, unit_x_multiplier, unit_y_multiplier, measure
        )
        if measure.minY > node.y:
            measure.minY = node.y

        if measure.maxY < node.y:
            measure.maxY = node.y

        if measure.minX > node.x:
            measure.minX = node.x

        if measure.maxX < node.x:
            measure.maxX = node.x

        measure.width = abs(measure.minX - measure.maxX)
        measure.height = abs(measure.minY - measure.maxY)
        measure.centerX = measure.minX + measure.width / 2
        measure.centerY = measure.minY + measure.height / 2
        return measure


class TidierExtreme:
    def __init__(self):
        self.left = None
        self.right = None
        self.thread = None
        self.offset = 0


class TreeMeasurement:
    """Summary of the rendered tree"""

    def __init__(self):
        self.minX = 10000
        self.maxX = 0
        self.minY = 10000
        self.maxY = 0
        self.width = 0
        self.height = 0
        self.centerX = 0
        self.centerY = 0

    def __repr__(self):
        return """tree measurement:
        bounds({}, {})
        min({}, {})
        max({}, {})
        center({}, {})""".format(
            self.width,
            self.height,
            self.minX,
            self.minY,
            self.maxX,
            self.maxY,
            self.centerX,
            self.centerY,
        )



import pytest

from ..mathy.agents.zero import SelfPlayConfig, self_play_runner


def test_self_play_runner_errors():
    """Throws error for bad profiling configuration of zero agent.
    
    This is only because it wasn't clear how to profile multiple processes that
    stop/start overtime during training/evaluation loops"""
    with pytest.raises(NotImplementedError):
        self_play_runner(SelfPlayConfig(profile=True, num_workers=2))



"""Some helper functions for PyTorch, including:
    - get_mean_and_std: calculate the mean and std value of dataset.
    - msr_init: net parameter initialization.
    - progress_bar: progress bar mimic xlua.progress.
"""
import errno
import os
import sys
import time
import math
from multiprocessing import Value

__all__ = ["AverageMeter"]


class AverageMeter(object):
    """Computes and stores the average and current value
       Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """

    def __init__(self):
        self.reset()

    @property
    def val(self):
        return self._val.value

    @property
    def sum(self):
        return self._sum.value

    @property
    def count(self):
        return self._count.value

    @property
    def avg(self):
        return self._avg.value

    def reset(self):
        self._val = Value("d", 0)
        self._sum = Value("d", 0)
        self._count = Value("d", 0)
        self._avg = Value("d", 0)

    def update(self, val, n=1):
        self._val.value = val
        self._sum.value += val * n
        self._count.value += n
        self._avg.value = self.sum / self.count



import os
import pickle
import time
from pathlib import Path
from shutil import copyfile
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import numpy as np
import srsly
import tensorflow as tf
from tensorflow.keras import backend as K
from wasabi import msg

from mathy.agents.example import example

from ..env import MathyEnv
from ..envs import PolySimplify
from ..state import (
    MathyInputsType,
    MathyObservation,
    MathyWindowObservation,
    ObservationFeatureIndices,
    observations_to_window,
)
from ..util import print_error
from .base_config import BaseConfig
from .embedding import MathyEmbedding


class PolicyValueModel(tf.keras.Model):
    args: BaseConfig
    optimizer: tf.optimizers.Optimizer

    def __init__(
        self,
        args: BaseConfig = None,
        predictions=2,
        initial_state: Any = None,
        **kwargs,
    ):
        super(PolicyValueModel, self).__init__(**kwargs)
        if args is None:
            args = BaseConfig()
        self.optimizer = tf.keras.optimizers.Adam(lr=args.lr)
        self.args = args
        self.predictions = predictions
        self.embedding = MathyEmbedding(self.args)
        self.value_head = tf.keras.layers.Dense(
            self.args.units,
            name="value_head",
            kernel_initializer="he_normal",
            activation="relu",
        )
        self.value_logits = tf.keras.layers.Dense(
            1, name="value_logits", kernel_initializer="he_normal", activation=None,
        )
        self.policy_td_dense = tf.keras.layers.Dense(
            self.predictions,
            name="timestep_dense",
            kernel_initializer="he_normal",
            activation=None,
        )
        self.policy_logits = tf.keras.layers.TimeDistributed(
            self.policy_td_dense, name="policy_logits"
        )
        self.normalize_pi = tf.keras.layers.LayerNormalization(name="policy_layer_norm")
        self.normalize_v = tf.keras.layers.LayerNormalization(name="value_layer_norm")
        self.loss = tf.Variable(
            0.0, trainable=False, name="loss_placeholder", dtype=tf.float32
        )

    def compute_output_shape(
        self, input_shapes: List[tf.TensorShape]
    ) -> tf.TensorShape:
        nodes_shape: tf.TensorShape = input_shapes[0]
        return tf.TensorShape(
            (
                nodes_shape.dims[0].value,
                nodes_shape.dims[1].value,
                self.args.embedding_units,
            )
        )

    def call(
        self, features_window: MathyInputsType
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return self._call(features_window)

    # @profile
    def _call(
        self, features_window: MathyInputsType
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        call_print = self.args.print_model_call_times
        nodes = features_window[ObservationFeatureIndices.nodes]
        batch_size = (
            len(nodes)
            if not isinstance(nodes, (tf.Tensor, np.ndarray))
            else nodes.shape[0]
        )
        start = time.time()
        inputs = features_window
        # Extract features into contextual inputs, sequence inputs.
        sequence_inputs = self.embedding(inputs)
        values = self.value_head(tf.reduce_mean(sequence_inputs, axis=1))
        values = self.normalize_v(values)
        values = self.value_logits(values)
        logits = self.normalize_pi(self.policy_logits(sequence_inputs))
        mask_logits = self.apply_pi_mask(logits, features_window)
        if call_print is True:
            print(
                "call took : {0:03f} for batch {1}".format(
                    time.time() - start, batch_size
                )
            )
        return logits, values, mask_logits

    def apply_pi_mask(
        self, logits: tf.Tensor, features_window: MathyInputsType,
    ) -> tf.Tensor:
        """Take the policy_mask from a batch of features and multiply
        the policy logits by it to remove any invalid moves"""
        mask = features_window[ObservationFeatureIndices.mask]
        logits_shape = tf.shape(logits)
        features_mask = tf.reshape(
            mask, (logits_shape[0], -1, self.predictions), name="pi_mask_reshape"
        )
        features_mask = tf.cast(features_mask, dtype=tf.float32)
        mask_logits = tf.multiply(logits, features_mask, name="mask_logits")
        negative_mask_logits = tf.where(
            tf.equal(mask_logits, tf.constant(0.0)),
            tf.fill(tf.shape(logits), -1000000.0),
            mask_logits,
            name="softmax_negative_logits",
        )
        return negative_mask_logits

    def predict_next(self, inputs: MathyInputsType) -> Tuple[tf.Tensor, tf.Tensor]:
        """Predict one probability distribution and value for the
        given sequence of inputs """
        logits, values, masked = self.call(inputs)
        # take the last timestep
        masked = masked[-1][:]
        flat_logits = tf.reshape(tf.squeeze(masked), [-1])
        probs = tf.nn.softmax(flat_logits)
        return probs, tf.squeeze(values[-1])

    def save(self) -> None:
        if not os.path.exists(self.args.model_dir):
            os.makedirs(self.args.model_dir)
        model_path = os.path.join(self.args.model_dir, self.args.model_name)
        with open(f"{model_path}.optimizer", "wb") as f:
            pickle.dump(self.optimizer.get_weights(), f)
        model_path += ".h5"
        self.save_weights(model_path, save_format="keras")
        step = self.optimizer.iterations.numpy()
        print(f"[save] step({step}) model({model_path})")


def _load_model(
    model: PolicyValueModel, model_file: str, optimizer_file: str,
) -> PolicyValueModel:
    model.load_weights(model_file)
    model._make_train_function()
    with open(optimizer_file, "rb") as f:
        weight_values = pickle.load(f)
    model.optimizer.set_weights(weight_values)
    return model


def get_or_create_policy_model(
    args: BaseConfig,
    predictions: int,
    is_main=False,
    required=False,
    env: MathyEnv = None,
) -> PolicyValueModel:
    if env is None:
        env = PolySimplify()
    observation: MathyObservation = env.state_to_observation(
        env.get_initial_state()[0], rnn_size=args.lstm_units
    )
    initial_state: MathyWindowObservation = observations_to_window([observation])

    if not os.path.exists(args.model_dir):
        os.makedirs(args.model_dir)
    model_path = os.path.join(args.model_dir, args.model_name)
    # Transfer weights/optimizer from a different model
    if is_main and args.init_model_from is not None:
        init_model_path = os.path.join(args.init_model_from, args.model_name)
        opt = f"{init_model_path}.optimizer"
        mod = f"{init_model_path}.h5"
        cfg = f"{init_model_path}.config.json"
        if os.path.exists(f"{model_path}.bytes"):
            print_error(
                ValueError("Model Exists"),
                f"Cannot initialize on top of model: {model_path}",
                print_error=False,
            )
        if os.path.exists(opt) and os.path.exists(mod) and os.path.exists(cfg):
            print(f"initialize model from: {init_model_path}")
            copyfile(opt, f"{model_path}.optimizer")
            copyfile(mod, f"{model_path}.h5")
            copyfile(cfg, f"{model_path}.config.json")
        else:
            print_error(
                ValueError("Model Exists"),
                f"Cannot initialize on top of model: {model_path}",
                print_error=False,
            )

    model = PolicyValueModel(args=args, predictions=predictions, name="agent")
    init_inputs = initial_state.to_inputs()
    model.compile(
        optimizer=model.optimizer, loss="binary_crossentropy", metrics=["accuracy"]
    )
    model.build(initial_state.to_input_shapes())
    model.predict(init_inputs)
    model.predict_next(init_inputs)

    opt = f"{model_path}.optimizer"
    mod = f"{model_path}.h5"
    if os.path.exists(mod):
        if is_main and args.verbose:
            with msg.loading(f"Loading model: {mod}..."):
                _load_model(model, mod, opt)
            msg.good(f"Loaded model: {mod}")
        else:
            _load_model(model, mod, opt)

        # If we're doing transfer, reset optimizer steps
        if is_main and args.init_model_from is not None:
            msg.info("reset optimizer steps to 0 for transfer model")
            model.optimizer.iterations.assign(0)
    elif required:
        print_error(
            ValueError("Model Not Found"),
            f"Cannot find model: {mod}",
            print_error=False,
        )
    elif is_main:
        cfg = f"{model_path}.config.json"
        if args.verbose:
            msg.info(f"wrote model config: {cfg}")
        srsly.write_json(cfg, args.dict(exclude_defaults=False))

    return model


def load_policy_value_model(
    model_data_folder: str, silent: bool = False
) -> PolicyValueModel:
    meta_file = Path(model_data_folder) / "model.config.json"
    if not meta_file.exists():
        raise ValueError(f"model meta not found: {meta_file}")
    args = BaseConfig(**srsly.read_json(str(meta_file)))
    model_file = Path(model_data_folder) / "model.h5"
    optimizer_file = Path(model_data_folder) / "model.optimizer"
    if not model_file.exists():
        raise ValueError(f"model not found: {model_file}")
    if not optimizer_file.exists():
        raise ValueError(f"optimizer not found: {optimizer_file}")
    env: MathyEnv = PolySimplify()
    observation: MathyObservation = env.state_to_observation(
        env.get_initial_state()[0], rnn_size=args.lstm_units
    )
    initial_state: MathyWindowObservation = observations_to_window([observation])
    model = PolicyValueModel(args=args, predictions=env.action_size, name="agent")
    init_inputs = initial_state.to_inputs()
    model.compile(
        optimizer=model.optimizer, loss="binary_crossentropy", metrics=["accuracy"]
    )
    model.build(initial_state.to_input_shapes())
    model.predict(init_inputs)
    model.predict_next(init_inputs)

    if not silent:
        with msg.loading(f"Loading model: {model_file}..."):
            _load_model(model, str(model_file), str(optimizer_file))
        msg.good(f"Loaded model: {model_file}")
    else:
        _load_model(model, str(model_file), str(optimizer_file))
    return model, args



from pathlib import Path
from setuptools import setup, find_packages


def setup_package():

    package_name = "mathy"
    root = Path(__file__).parent.resolve()

    # Read in package meta from about.py
    about_path = root / package_name / "about.py"
    with about_path.open("r", encoding="utf8") as f:
        about = {}
        exec(f.read(), about)

    with open(root / "README.md", "r") as fh:
        long_description = fh.read()

    with open(root / "requirements.txt") as file:
        REQUIRED_MODULES = [line.strip() for line in file]

    with open(root / "requirements-dev.txt") as file:
        DEVELOPMENT_MODULES = [line.strip() for line in file if "-e" not in line]

    setup(
        name=package_name,
        description=about["__summary__"],
        author=about["__author__"],
        author_email=about["__email__"],
        url=about["__uri__"],
        version=about["__version__"],
        license=about["__license__"],
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords="math",
        install_requires=REQUIRED_MODULES,
        packages=find_packages(),
        extras_require={"dev": DEVELOPMENT_MODULES},
        package_data={"mathy": ["tests/api/*.json", "tests/rules/*.json"]},
        entry_points="""
            [console_scripts]
            mathy=mathy.cli:cli
        """,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Development Status :: 2 - Pre-Alpha",
        ],
        python_requires=">=3.6",
        include_package_data=True,
    )


if __name__ == "__main__":
    setup_package()



from multiprocessing import Queue
from typing import Dict, List, Optional

import numpy as np

from ..state import (
    MathyObservation,
    MathyWindowObservation,
    observations_to_window,
    rnn_placeholder_state,
)


def rnn_weighted_history(
    observations: List[MathyObservation], rnn_size: int = 128,
):
    """Build a historical LSTM state: https://arxiv.org/pdf/1810.04437.pdf

    Take the mean of the previous LSTM states for this episode. """

    import tensorflow as tf

    if len(observations) > 0:
        in_h = []
        in_c = []
        for obs in observations:
            in_h.append(obs.rnn_state_h)
            in_c.append(obs.rnn_state_c)
        # Take the mean of the historical states:
        memory_context_h = tf.reduce_mean(in_h, axis=0)
        memory_context_c = tf.reduce_mean(in_c, axis=0)
    else:
        memory_context_h = rnn_placeholder_state(rnn_size)
        memory_context_c = memory_context_h
    return [memory_context_h, memory_context_c]


class EpisodeMemory(object):
    # Observation from the environment
    observations: List[MathyObservation]
    # Action taken in the environment
    actions: List[int]
    # Reward from the environmnet
    rewards: List[float]
    # Estimated value from the model
    values: List[float]
    # Grouping Control error from the environment
    grouping_changes: List[float]

    def __init__(self):
        self.clear()

    def clear(self):
        self.observations = []
        self.actions = []
        self.rewards = []
        self.values = []
        self.grouping_changes = []

    def to_window_observation(
        self, observation: MathyObservation, window_size: int = 3
    ) -> MathyWindowObservation:
        previous = -(max(window_size - 1, 1))
        window_observations = self.observations[previous:] + [observation]
        return observations_to_window(window_observations)

    def to_episode_window(self) -> MathyWindowObservation:
        return observations_to_window(self.observations)

    def store(
        self,
        *,
        observation: MathyObservation,
        action: int,
        reward: float,
        grouping_change: float,
        value: float,
    ):
        self.observations.append(observation)
        self.actions.append(action)
        self.rewards.append(reward)
        self.values.append(value)
        self.grouping_changes.append(grouping_change)

    def rnn_weighted_history(self, rnn_size):
        return rnn_weighted_history(self.observations, rnn_size)



import math
from typing import List, Optional, Type

import gym
import numpy as np
from gym.envs.registration import register

from ...core.expressions import MathExpression
from ...core.rule import ExpressionChangeRule
from ...env import MathyEnv
from ...state import (
    MathyEnvState,
    MathyObservation,
    RNNStateFloatList,
    observations_to_window,
    rnn_placeholder_state,
)
from ...types import MathyEnvProblemArgs, MathyEnvProblem
from ...util import is_terminal_transition
from .masked_discrete import MaskedDiscrete


class MathyGymEnv(gym.Env):
    """A small wrapper around Mathy envs to allow them to work with OpenAI Gym. The
    agents currently use this env wrapper, but it could be dropped in the future."""

    default_rnn_size = 128
    metadata = {"render.modes": ["terminal", "attention"]}
    mathy: MathyEnv
    state: Optional[MathyEnvState]
    problem: Optional[MathyEnvProblem]
    env_class: Type[MathyEnv]
    env_problem_args: Optional[MathyEnvProblemArgs]
    last_action: int
    last_reward: float
    last_change: Optional[ExpressionChangeRule]

    def __init__(
        self,
        env_class: Type[MathyEnv] = MathyEnv,
        env_problem_args: Optional[MathyEnvProblemArgs] = None,
        **env_kwargs,
    ):
        self.mathy = env_class(**env_kwargs)
        self.env_class = env_class
        self.env_problem_args = env_problem_args
        if self.env_problem_args is not None and not isinstance(
            self.env_problem_args, MathyEnvProblemArgs
        ):
            raise ValueError("Problem args must be a MathyEnvProblemArgs instance")

        self.last_action = -1
        self.last_change = None
        self.last_reward = 0.0
        max_nodes = 1024
        max_actions = self.mathy.action_size
        vector_width = 1  # a single number
        self.state = None
        self.problem = None
        self.vectors_shape = (max_nodes, vector_width)
        self.action_space = MaskedDiscrete(max_actions, [1] * max_actions)

    @property
    def action_size(self) -> int:
        if self.state is not None:
            return self.mathy.get_agent_actions_count(self.state)
        return self.mathy.action_size

    def step(self, action):
        self.state, transition, change = self.mathy.get_next_state(self.state, action)
        observation = self._observe(self.state)
        info = {"transition": transition}
        done = is_terminal_transition(transition)
        self.last_action = action
        self.last_change = change
        self.last_reward = round(float(transition.reward), 4)
        return observation, transition.reward, done, info

    def reset_with_input(
        self, problem_text: str, rnn_size=default_rnn_size, max_moves=16
    ):
        self.rnn_size = rnn_size
        self.last_action = -1
        self.last_change = None
        self.last_reward = 0.0
        # If the episode is being reset because it ended, assert the validity
        # of the last problem outcome
        if self.state is not None:
            self.mathy.finalize_state(self.state)
        self.state, self.problem = self.mathy.get_initial_state(self.env_problem_args)
        self.state = MathyEnvState(problem=problem_text, max_moves=max_moves)
        return self._observe(self.state)

    def reset(self, rnn_size=default_rnn_size):
        self.rnn_size = rnn_size
        self.last_action = -1
        self.last_change = None
        self.last_reward = 0.0
        # If the episode is being reset because it ended, assert the validity
        # of the last problem outcome
        if self.state is not None:
            self.mathy.finalize_state(self.state)
        self.state, self.problem = self.mathy.get_initial_state(self.env_problem_args)
        return self._observe(self.state)

    def initial_window(self, rnn_size: int):
        """return an n-step set of observations for initializing the env"""
        state, _ = self.mathy.get_initial_state(self.env_problem_args)
        return observations_to_window(
            [self.mathy.state_to_observation(state, rnn_size=rnn_size)]
        )

    def _observe(self, state: MathyEnvState) -> MathyObservation:
        """Observe the environment at the given state, updating the observation
        space and action space for the given state. """
        action_mask = self.mathy.get_valid_moves(state)
        observation = self.mathy.state_to_observation(state, rnn_size=self.rnn_size)
        self.action_space.n = self.mathy.get_agent_actions_count(state)
        self.action_space.mask = action_mask
        return observation

    def render(self, mode="terminal", data=None):
        action_name = "initial"
        token_index = -1
        if self.last_action != -1 and self.last_change is not None:
            action_index, token_index = self.mathy.get_action_indices(self.last_action)
            action_name = self.mathy.rules[action_index].name
        else:
            print(f"Problem: {self.state.agent.problem}")
        self.mathy.print_state(
            self.state,
            action_name[:25].lower(),
            token_index,
            change=self.last_change,
            change_reward=self.last_reward,
        )


def safe_register(id: str, **kwargs):
    """Ignore re-register errors."""
    try:
        return register(id, **kwargs)
    except gym.error.Error:
        pass



import os
import numpy as np
from . import time_step
from pydantic import BaseModel
import json
import math
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Union, Any
import re
from typing import Dict, List, Tuple
from wasabi import TracebackPrinter

import numpy as np

from .types import EnvRewards, Literal

from .core.expressions import (
    AddExpression,
    BinaryExpression,
    ConstantExpression,
    DivideExpression,
    MathExpression,
    MultiplyExpression,
    NegateExpression,
    PowerExpression,
    SubtractExpression,
    VariableExpression,
)
from .core.parser import ExpressionParser
from .core.tree import LEFT


def is_debug_mode():
    """Debug mode enables extra logging and assertions, but is slower because of
    the increased sanity check measurements."""
    return False


def compare_expression_string_values(
    from_expression: str, to_expression: str, history: Optional[List[Any]] = None
):
    """Compare and evaluate two expressions strings to verify they have the
    same value"""
    parser = ExpressionParser()
    return compare_expression_values(
        parser.parse(from_expression), parser.parse(to_expression), history
    )


def raise_with_history(
    title: str, description: str, history: Optional[List[Any]] = None,
):
    import traceback

    history_text: List[str] = []
    if history is not None:
        history_text = [f" - {h.raw}" for h in history]
    history_text.insert(0, description)
    tb = TracebackPrinter()
    error = tb(title, "\n".join(history_text), tb=traceback.extract_stack())
    raise ValueError(error)


def compare_expression_values(
    from_expression: MathExpression,
    to_expression: MathExpression,
    history: Optional[List[Any]] = None,
):
    """Compare and evaluate two expressions to verify they have the same value"""
    vars_from: set = set()
    vars_to: set = set()
    for v in from_expression.find_type(VariableExpression):
        vars_from.add(v.identifier)
    for v in to_expression.find_type(VariableExpression):
        vars_to.add(v.identifier)
    # If there are not the same unique vars in the two expressions, something
    # bad happened, and the two expressions can only coincidentally be equal
    # in value.
    if len(vars_from) != len(vars_to):
        raise_with_history(
            "Number of variables changed",
            f"{list(vars_from)} != {list(vars_to)}",
            history,
        )

    sorted_from = list(vars_from)
    sorted_from.sort()
    sorted_to = list(vars_from)
    sorted_to.sort()

    if sorted_from != sorted_to:
        raise_with_history(
            "Unique variables changed", f"{sorted_from} != {sorted_to}", history,
        )

    # Generate random values for each variable, and then evaluate the expressions
    eval_context: Dict[str, float] = {}
    for var in vars_from:
        eval_context[var] = float(np.random.randint(1, 100))

    value_from = from_expression.evaluate(eval_context)
    value_to = to_expression.evaluate(eval_context)

    # Print out the problem steps leading up to error result.
    if not math.isclose(value_from, value_to, rel_tol=1e-9, abs_tol=0.0):
        changed = f"""
        {from_expression} = {value_from}

        {to_expression} = {value_to}

        {value_from} != {value_to}
        """
        raise_with_history("Expression value changed", changed, history)


def unlink(node: Optional[MathExpression] = None) -> Optional[MathExpression]:
    """Unlink an expression from it's parent.

      1. Clear expression references in `parent`
      2. Clear `parent` in expression
    """

    if node is None:
        return None

    if node.parent is not None and node == node.parent.left:
        node.parent.set_left(None)

    if node.parent is not None and node == node.parent.right:
        node.parent.set_right(None)

    node.parent = None
    return node


# Build a verbose factor dictionary.
#
# This builds a dictionary of factors for a given value that
# contains both arrangements of terms so that all factors are
# accessible by key.  That is, factoring 2 would return
#      result =
#        1 : 2
#        2 : 1
def factor(value) -> Dict[int, int]:
    if value == 0 or math.isnan(value):
        return {}
    np.seterr(invalid="ignore")
    sqrt = np.sqrt(value)
    np.seterr(invalid="warn")
    if math.isnan(sqrt):
        return {1: value}

    sqrt = int(sqrt + 1)
    factors = {1: value}
    factors[value] = 1

    for i in range(2, sqrt):
        if value % i == 0:
            one = i
            two = value / i
            factors[one] = two
            factors[two] = one
    return factors


def is_add_or_sub(node):
    return isinstance(node, AddExpression) or isinstance(node, SubtractExpression)


def get_sub_terms(node: MathExpression):
    nodes = node.to_list("inorder")
    terms = []

    def safe_pop():
        nonlocal nodes
        if len(nodes) > 0:
            return nodes.pop(0)
        return None

    current = safe_pop()

    while current is not None:
        term_const = term_var = term_exp = None
        # If there's a coefficient, note it
        if isinstance(current, ConstantExpression):
            term_const = current
            current = safe_pop()
            # Cannot have add/sub in sub-terms
            if is_add_or_sub(current):
                return False
            # It should be one of these operators
            assert current is None or isinstance(
                current, (MultiplyExpression, DivideExpression, PowerExpression)
            )
            if not isinstance(current, PowerExpression):
                current = safe_pop()

        # Pop off the variable
        if isinstance(current, VariableExpression):
            term_var = current
            current = safe_pop()
            # cannot have add/sub in sub-terms
            if is_add_or_sub(current):
                return False
            # It should be one of these operators
            assert current is None or isinstance(
                current, (MultiplyExpression, DivideExpression, PowerExpression)
            )
            if not isinstance(current, PowerExpression):
                current = safe_pop()

        # Look for exponent
        if isinstance(current, PowerExpression):
            # pop the power binary expression
            current = safe_pop()
            # store the right hand side
            term_exp = current
            # pop it off for next term
            current = safe_pop()

        # Couldn't find anything
        if term_const is None and term_exp is None and term_var is None:
            # If it's a continuation multiply, e.g in "4 * (x^2 * z^6)"
            # NOTE: it's a continuation mult in the above expression
            #   because of the parens grouping between the 4 multiply
            #   and the inner multiply of the parens sub-terms
            if isinstance(current, MultiplyExpression):
                # pop it off for next term
                current = safe_pop()
                continue

            return False
        terms.append((term_const, term_var, term_exp))
    return terms


def is_simple_term(node: MathExpression) -> bool:
    """Return True if a given term has been simplified so it only has at
    most one of each variable and a constant.

    # Examples
      - Simple = 2x^2 * 2y
      - Complex = 2x * 2x * 2y
      - Simple = x^2 * 4
      - Complex = 2 * 2x^2
    """
    sub_terms = get_sub_terms(node)
    if sub_terms is False:
        return False
    seen: set = set()
    co_key = "coefficient"

    for coefficient, variable, exponent in sub_terms:
        if coefficient is not None:
            if co_key in seen:
                return False
            seen.add(co_key)

        if variable is not None or exponent is not None:
            key = "{}{}".format(variable, exponent)
            if key in seen:
                return False
            seen.add(key)

    return True


def is_preferred_term_form(expression: MathExpression) -> bool:
    """
    Return True if a given term has been simplified so that it only has
    a max of one coefficient and variable, with the variable on the right
    and the coefficient on the left side

    Examples
    
      - Complex   = 2 * 2x^2
      - Simple    = x^2 * 4
      - Preferred = 4x^2
    """
    if not is_simple_term(expression):
        return False

    # If there are multiple multiplications this term can be simplified further. At most
    # we expect a multiply to connect a coefficient and variable.
    # NOTE: the following check is removed because we need to handle multiple variable
    #       terms, e.g. "4x * z"
    # if len(expression.find_type(MultiplyExpression)) > 1:
    #     return False

    # if there's a variable, make sure the coefficient is on the left side
    # for the preferred compact form. i.e. "4x" instead of "x * 4"
    vars: List[VariableExpression] = expression.find_type(VariableExpression)
    seen_vars: Dict[str, int] = dict()
    parent: MathExpression
    for var in vars:
        if var.identifier is None:
            continue
        if var.identifier not in seen_vars:
            seen_vars[var.identifier] = 0
        seen_vars[var.identifier] += 1
        parent = var
        if isinstance(var.parent, PowerExpression):
            parent = var.parent
        if parent.parent is not None and parent.parent.get_side(parent) == LEFT:
            if isinstance(parent.parent.right, ConstantExpression):
                return False
    # If any variables show up more than once, they can be combined.
    # examples:
    #  - `b * (44b^2)`
    #  - `z * (1274z^2)`
    for key, value in seen_vars.items():
        if value > 1:
            return False
    return True


def has_like_terms(expression: MathExpression) -> bool:
    """Return True if a given expression has more than one of any type of term.

    # Examples

    - `x + y + z` = `False`
    - `x^2 + x` = `False`
    - `y + 2x` = `True`
    - `x^2 + 4x^3 + 2y` = `True`
    """

    seen: set = set()
    term_nodes = get_terms(expression)
    for node in term_nodes:
        term = get_term(node)
        if not isinstance(term, TermResult):
            continue
        var_key = ("".join(term.variables), term.exponent)
        # If the same var/power combinaton is found in the expression more than once
        # there are like terms.
        if var_key in seen:
            return True
        seen.add(var_key)

    # Look for multiple free-floating constants
    consts = expression.find_type(ConstantExpression)
    for const in consts:
        if const.parent and is_add_or_sub(const.parent):
            if "const_term" in seen:
                return True
            seen.add("const_term")
    return False


class FactorResult:
    def __init__(self):
        self.best = -1
        self.left = -1
        self.right = -1
        self.all_left = []
        self.all_right = []
        self.variable = None
        self.exponent = None
        self.leftExponent = None
        self.rightExponent = None
        self.leftVariable = None
        self.rightVariable = None


# Create a term node hierarchy from a given set of
# term parameters.  This takes into account removing
# implicit coefficients of 1 where possible.
def make_term(coefficient, variable, exponent):
    constExp = ConstantExpression(coefficient)
    if not variable and not exponent:
        return constExp

    varExp = VariableExpression(variable)
    if coefficient == 1 and not exponent:
        return varExp

    multExp = MultiplyExpression(constExp, varExp)
    if not exponent:
        return multExp

    expConstExp = ConstantExpression(exponent)
    if coefficient == 1:
        return PowerExpression(varExp, expConstExp)

    return PowerExpression(multExp, expConstExp)


class TermResult:
    def __init__(self):
        self.coefficients = []
        self.variables = []
        self.exponent = None
        self.node_coefficients = []
        self.node_variables = []
        self.node_exponent = None


# Extract term information from the given node
#
def get_term(node: MathExpression) -> Union[TermResult, Literal[False]]:
    result = TermResult()
    # Constant with add/sub parent should be OKAY.
    if isinstance(node, ConstantExpression):
        if not node.parent or (node.parent and is_add_or_sub(node.parent)):
            result.coefficients = [node.value]
            result.node_coefficients = [node]
            return result

    # Variable with add/sub parent should be OKAY.
    if isinstance(node, VariableExpression):
        if not node.parent or (node.parent and is_add_or_sub(node.parent)):
            result.variables = [node.identifier]
            result.node_variables = [node]
            return result

    # TODO: Comment resolution on whether +- is OKAY, and if not, why it breaks down.
    if not is_add_or_sub(node):
        if (
            len(node.find_type(AddExpression)) > 0
            or len(node.find_type(SubtractExpression)) > 0
        ):
            return False

    # If another add is found on the left side of this node, and the right node
    # is _NOT_ a leaf, we cannot extract a term.  If it is a leaf, the term should be
    # just the right node.
    if node.left and len(node.left.find_type(AddExpression)) > 0:
        if node.right and not node.right.is_leaf():
            return False

    if node.right and len(node.right.find_type(AddExpression)) > 0:
        return False

    exponents = node.find_type(PowerExpression)
    if len(exponents) > 0:
        # Supports only single exponents in terms
        if len(exponents) != 1:
            return False

        exponent = exponents[0]
        if not isinstance(exponent.right, ConstantExpression):
            return False

        result.exponent = exponent.right.value
        result.node_exponent = exponent

    variables = node.find_type(VariableExpression)
    if len(variables) > 0:
        result.variables = [v.identifier for v in variables]
        result.variables.sort()
        result.node_variables = variables

    def filter_coefficients(n):
        if not n.parent or n.parent == node.parent:
            return True

        if isinstance(n.parent, BinaryExpression) and not isinstance(
            n.parent, MultiplyExpression
        ):
            return False
        return True

    coefficients = node.find_type(ConstantExpression)
    coefficients = [c for c in coefficients if filter_coefficients(c)]
    if len(coefficients) > 0:

        def resolve_coefficients(c):
            value = c.value
            if isinstance(c.parent, NegateExpression):
                value = value * -1

            return value

        result.coefficients = [resolve_coefficients(c) for c in coefficients]
        result.node_coefficients = coefficients

    empty = (
        len(result.variables) == 0
        and len(result.coefficients) == 0
        and result.exponent is None
    )
    if empty:
        return False

    # consistently return an empty coefficients/variables array in case none exist.
    # this ensures that you can always reference coefficients[0] or variables[0] and
    # check that for truthiness, rather than having to check that the object property
    # `coefficients` or `variables` is not None and also the truthiness of index 0.
    if not result.coefficients:
        result.coefficients = []

    if not result.variables:
        result.variables = []
    return result


class TermEx(NamedTuple):
    coefficient: Optional[Union[int, float]]
    variable: Optional[str]
    exponent: Optional[Union[int, float]]


# fmt: off
TermEx.coefficient.__doc__ = "An optional integer or float coefficient" # noqa
TermEx.variable.__doc__ = "An optional variable" # noqa
TermEx.exponent.__doc__ = "An optional integer or float exponent" # noqa
# fmt: on


def get_term_ex(node: Optional[MathExpression]) -> Optional[TermEx]:
    """Extract the 3 components of a naturally ordered term.

    !!! info Important

        This doesn't care about whether the node is part of a larger term,
        it only looks at its children.

    # Example

    `mathy:4x^7`

    ```python
    TermEx(coefficient=4, variable="x", exponent=7)
    ```
    """

    # "4"
    if isinstance(node, ConstantExpression):
        # Make sure the parent isn't an exponent link (in which case we'd incorrectly
        # report this as a term with no exponent.) That case can be handled by calling
        # this on the parent node.
        if node.parent is None or not isinstance(node.parent, PowerExpression):
            return TermEx(node.value, None, None)

    # "x"
    if isinstance(node, VariableExpression):
        # Make sure the parent isn't an exponent link (in which case we'd incorrectly
        # report this as a term with no exponent.) That case can be handled by calling
        # this on the parent node.
        if node.parent is None or not isinstance(node.parent, PowerExpression):
            return TermEx(None, node.identifier, None)
    # "4 * ???"
    if isinstance(node, MultiplyExpression) and isinstance(
        node.left, ConstantExpression
    ):
        # "4x"
        if isinstance(node.right, VariableExpression):
            return TermEx(node.left.value, node.right.identifier, None)

        # "4x^2"
        if isinstance(node.right, PowerExpression):
            pow = node.right
            if isinstance(pow.left, VariableExpression) and isinstance(
                pow.right, ConstantExpression
            ):
                return TermEx(node.left.value, pow.left.identifier, pow.right.value)

    # "x^2"
    if isinstance(node, PowerExpression):
        if isinstance(node.left, VariableExpression) and isinstance(
            node.right, ConstantExpression
        ):
            return TermEx(None, node.left.identifier, node.right.value)

    return None


def factor_add_terms_ex(left_term: TermEx, right_term: TermEx) -> FactorResult:
    if not left_term or not right_term:
        raise ValueError("invalid terms for factoring")

    # Common coefficients
    l_factors = factor(
        left_term.coefficient if left_term.coefficient is not None else 1
    )
    r_factors = factor(
        right_term.coefficient if right_term.coefficient is not None else 1
    )
    common = [k for k in r_factors if k in l_factors]
    if len(common) == 0:
        return False
    has_left: bool = left_term.variable is not None
    has_right: bool = right_term.variable is not None

    # If there are variables, we want to extract them, so
    # the smallest number to factor out. TODO: is this okay?
    if has_left or has_right:
        best = np.min(common)
    else:
        best = np.max(common)
    result = FactorResult()
    result.best = best
    result.left = l_factors[best]
    result.right = r_factors[best]
    result.all_left = l_factors
    result.all_right = r_factors

    # Common variables and powers
    two_exp_and_match: bool = (
        left_term.exponent is not None
        and right_term.exponent is not None
        and left_term.exponent == right_term.exponent
    )
    both_match = (
        False
        if (left_term.exponent or right_term.exponent) and not two_exp_and_match
        else True
    )
    if (
        has_left
        and has_right
        and left_term.variable == right_term.variable
        and both_match
    ):
        result.variable = left_term.variable
        result.exponent = left_term.exponent

    if left_term.exponent and left_term.exponent != result.exponent:
        result.leftExponent = left_term.exponent

    if right_term.exponent and right_term.exponent != result.exponent:
        result.rightExponent = right_term.exponent

    if has_left and left_term.variable != result.variable:
        result.leftVariable = left_term.variable

    if has_right and right_term.variable != result.variable:
        result.rightVariable = right_term.variable

    return result


def get_terms(expression: MathExpression) -> List[MathExpression]:
    """Walk the given expression tree and return a list of nodes
    representing the distinct terms it contains.
    
    # Arguments
    expression (MathExpression): the expression to find term nodes in

    # Returns
    (List[MathExpression]): a list of term nodes
    """
    results: List[MathExpression] = []
    root = expression.get_root()
    if isinstance(root, MultiplyExpression):
        results.append(root)

    def visit_fn(node, depth, data):
        nonlocal results
        if not is_add_or_sub(node):
            return
        if not is_add_or_sub(node.left):
            results.append(node.left)
        if not is_add_or_sub(node.right):
            results.append(node.right)

    root.visit_inorder(visit_fn)
    return [expression] if len(results) == 0 else results


def terms_are_like(
    one: Union[TermResult, MathExpression, Literal[False]],
    two: Union[TermResult, MathExpression, Literal[False]],
) -> bool:
    """Determine if two math expression nodes are **like terms**.

    # Arguments
    one (MathExpression): A math expression that represents a term
    two (MathExpression): Another math expression that represents a term

    # Returns
    (bool): Whether the terms are like or not.
    """
    # Extract terms from MathExpressions if need be
    if isinstance(one, MathExpression):
        one = get_term(one)

    if isinstance(two, MathExpression):
        two = get_term(two)

    if one is False or two is False:
        return False

    # Coerce the type away from math expression
    assert isinstance(one, TermResult) and isinstance(two, TermResult)

    # if neither have variables, then they are a match!
    if len(one.variables) == 0 and len(two.variables) == 0:
        return True

    # Both must have variables, and they must match exactly.
    if len(one.variables) != len(two.variables):
        return False

    invalid = len([False for v in one.variables if v not in two.variables]) > 0
    if invalid:
        return False

    # Also, the exponents must match
    if one.exponent != two.exponent:
        return False

    # Same variables, and exponents.  Good to go.
    return True


def is_terminal_transition(transition: time_step.TimeStep) -> bool:
    return bool(transition.step_type == time_step.StepType.LAST)


def discount(values: List[float], gamma=0.99) -> List[float]:
    """Discount a list of floating point values.

    # Arguments
    r (List[float]): the list of floating point values to discount
    gamma (float): a value between 0 and 0.99 to use when discounting the inputs
    
    # Returns
    (List[float]): a list of the same size as the input with discounted values
    """
    discounted_r = np.zeros_like(values, dtype=np.float32)
    running_add = 0
    for t in reversed(range(len(values))):
        running_add = running_add * gamma + values[t]
        discounted_r[t] = running_add
    # reverse them to restore the correct order
    np.flip(discounted_r)
    return discounted_r


def pad_array(in_list: List[Any], max_length: int, value: Any = 0) -> List[Any]:
    """Pad a list to the given size with the given padding value.
    
    # Arguments:
    in_list (List[Any]): List of values to pad to the given length
    max_length (int): The desired length of the array
    value (Any): a value to insert in order to pad the array to max length

    # Returns
    (List[Any]): An array padded to `max_length` size
    """
    while len(in_list) < max_length:
        in_list.append(value)
    return in_list


def calculate_term_grouping_distances(input: str) -> Tuple[float, float]:
    vars = re.findall(r"([a-zA-Z]\^?\d?)+", input)
    seen_pos: Dict[str, List[int]] = dict()
    for i, var in enumerate(vars):
        if var not in seen_pos:
            seen_pos[var] = []
        seen_pos[var].append(i)

    def get_var_signal(var, positions):
        out_signal = 0.0
        last_pos = -1
        for position in positions:
            if last_pos != -1:
                out_signal += position - last_pos
            last_pos = position
        return out_signal

    # seen_pos is a dict of positions that each variable is seen at
    # add up all the distances between the points in each variable
    signal = 0.0
    for key, value in seen_pos.items():
        signal += get_var_signal(key, value)

    # Scale the signal down to avoid it growing ever larger with more
    # complex inputs.
    return signal, signal / len(vars)


def calculate_grouping_control_signal(
    input: str, output: str, clip_at_zero: bool = False
) -> float:
    """Calculate grouping_control signals as the sum of all distances between
    all like terms. Gather all the terms in an expression and add an error value
    whenever a like term is separated by another term."""

    # We cheat the term grouping a bit by not parsing the expression
    # and finding the real set of terms. Instead we remove all the non-variables
    # and then count the distances from the resulting string.
    #
    # NOTE: this means that the signal is not correct when exponents or complex
    #       terms with multiple variables are in the expression. Perhaps it's a
    #       good improvement to make here.
    in_signal, in_signal_normalized = calculate_term_grouping_distances(input)
    out_signal, out_signal_normalized = calculate_term_grouping_distances(output)
    # The grouping distance stayed the same
    if in_signal == out_signal:
        return out_signal_normalized

    # It changed, no error
    if clip_at_zero is True:
        return 0.0

    # It changed, negative error based on magnitude of the change
    return -abs(in_signal_normalized - out_signal_normalized)


def print_error(error, text, print_error=True):
    import traceback

    caught_error = TracebackPrinter(
        color_error="yellow", tb_base=".", tb_range_start=-15
    )(
        error.__class__.__name__,
        f"{error}",
        tb=traceback.extract_tb(error.__traceback__),
    )
    caught_at = TracebackPrinter(tb_base=".", tb_range_start=-15, tb_range_end=-1)(
        f"Error: {text}", f"Caught at:", tb=traceback.extract_stack(),
    )

    if print_error:
        print(caught_at + caught_error)
    else:
        raise ValueError(caught_at + caught_error)




# coding=utf-8
# Copyright 2018 The TF-Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Changes:
#
#  - 2019/12/09 JD: inline and remove tensorflow dependency/shape support
#  - 2019/12/11 JD: add MathyObservation type hints
"""TimeStep representing a step in the environment.

This file is a mostly direct copy of the implementation from the
[tf_agents](https://github.com/tensorflow/agents) library but has
the dependency on tensorflow removed along with advanced shape
features.

Mathy doesn't use these features and the overhead of loading tensorflow
to pass environment states around is not great for things like CLI start
times.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import numpy as np


class TimeStep(
    collections.namedtuple(
        "TimeStep", ["step_type", "reward", "discount", "observation"]
    )
):
    __slots__ = ()

    def is_first(self):
        return np.equal(self.step_type, StepType.FIRST)

    def is_mid(self):
        return np.equal(self.step_type, StepType.MID)

    def is_last(self):
        return np.equal(self.step_type, StepType.LAST)

    def __hash__(self):
        return hash(tuple(self))


class StepType(object):
    """Defines the status of a `TimeStep` within a sequence."""

    # Denotes the first `TimeStep` in a sequence.
    FIRST = np.asarray(0, dtype=np.int32)
    # Denotes any `TimeStep` in a sequence that is not FIRST or LAST.
    MID = np.asarray(1, dtype=np.int32)
    # Denotes the last `TimeStep` in a sequence.
    LAST = np.asarray(2, dtype=np.int32)

    def __new__(cls, value):
        """Add ability to create StepType constants from a value."""
        if value == cls.FIRST:
            return cls.FIRST
        if value == cls.MID:
            return cls.MID
        if value == cls.LAST:
            return cls.LAST

        raise ValueError("No known conversion for `%r` into a StepType" % value)


def transition(observation, reward, discount=1.0):
    """Returns a `TimeStep` with `step_type` set equal to `StepType.MID`."""
    return TimeStep(StepType.MID, reward, discount, observation)


def termination(observation, reward):
    """Returns a `TimeStep` with `step_type` set to `StepType.LAST`."""
    return TimeStep(StepType.LAST, reward, 00, observation)


def truncation(observation, reward, discount=1.0):
    """Returns a `TimeStep` with `step_type` set to `StepType.LAST`."""
    return TimeStep(StepType.LAST, reward, discount, observation)



from typing import Tuple
from ..mathy.agents.episode_memory import EpisodeMemory
from ..mathy.envs.poly_simplify import PolySimplify
from numpy.testing import assert_array_almost_equal


def get_memory(
    number_observations: int, rnn_size: int = 4
) -> Tuple[EpisodeMemory, PolySimplify]:
    """Verify that RNN history is the average of the RNN states in memory"""
    memory = EpisodeMemory()
    env = PolySimplify()
    state, problem = env.get_initial_state()
    for i in range(number_observations):
        memory.store(
            observation=env.state_to_observation(state, rnn_size=rnn_size),
            action=0,
            reward=0.0,
            grouping_change=0.0,
            value=0.0,
        )
    return memory, env


def test_episode_memory_to_episode_window():
    """verify that to_episode_window returns a window of the entire episode"""
    memory, env = get_memory(12)

    assert len(memory.observations) == 12

    assert len(memory.to_episode_window().nodes) == 12


def test_episode_memory_to_window_observation():
    """Verify that to_window_observation only returns a subset of the memory"""
    memory, env = get_memory(4)
    assert len(memory.observations) == 4
    obs = env.state_to_observation(env.get_initial_state()[0], rnn_size=4)
    # Returns only the last three
    assert len(memory.to_window_observation(obs, window_size=3).nodes) == 3
    # Only returns as many observations as it has (plus the one provided)
    assert len(memory.to_window_observation(obs, window_size=10).nodes) == 5


def test_episode_memory_rnn_history():
    """Verify that RNN history is the average of the RNN states in memory"""

    memory = EpisodeMemory()
    env = PolySimplify()
    state, problem = env.get_initial_state()
    # Two RNN states
    rnn_states = [
        [[0.0, 1.0, 0.8], [1.2, 0.2, 1.8]],
        [[0.5, 1.0, 0.4], [0.0, 2.2, 0.4]],
    ]
    # The expected element-wise average of those states
    rnn_assert = [[0.25, 1.0, 0.6], [0.6, 1.2, 1.1]]

    # Insert the RNN states into memory
    for r_state in rnn_states:
        memory.store(
            observation=env.state_to_observation(
                state, rnn_state_h=r_state[0], rnn_state_c=r_state[1]
            ),
            action=0.0,
            reward=0.0,
            grouping_change=0.0,
            value=0.0,
        )
    # Fetch and check the weighted RNN history value
    weighted_rnn = memory.rnn_weighted_history(len(rnn_assert[0]))
    assert_array_almost_equal(weighted_rnn, rnn_assert)



from ..mathy.state import MathyEnvState
from ..mathy.env import MathyEnv
from ..mathy.envs.poly_simplify import PolySimplify
from ..mathy.util import is_terminal_transition
import random


def test_mathy_env_jd():
    env = MathyEnv()
    assert env is not None
    problem = "5y * 9x + 8z + 8x + 3z * 10y * 11x + 10y"
    env_state = MathyEnvState(problem=problem, max_moves=35)
    for i in range(3):
        actions = env.get_valid_moves(env_state)
        indices = [i for i, value in enumerate(actions) if value == 1]
        random.shuffle(indices)
        env_state, value, changed = env.get_next_state(env_state, indices[0])
    assert env_state.to_observation([]) is not None


def test_mathy_env_win_conditions():

    expectations = [
        ("70656 * (x^2 * z^6)", True),
        ("b * (44b^2)", False),
        ("z * (1274z^2)", False),
        ("4x^2", True),
        ("100y * x + 2", True),
        ("10y * 10x + 2", False),
        ("10y + 1000y * (y * z)", False),
        ("4 * (5y + 2)", False),
        ("2", True),
        ("4x * 2", False),
        ("4x * 2x", False),
        ("4x + 2x", False),
        ("4 + 2", False),
        ("3x + 2y + 7", True),
        ("3x^2 + 2x + 7", True),
        ("3x^2 + 2x^2 + 7", False),
    ]

    # Valid solutions but out of scope so they aren't counted as wins.
    #
    # This works because the problem sets exclude this type of > 2 term
    # polynomial expressions
    out_of_scope_valid = []

    env = PolySimplify()
    for text, is_win in expectations + out_of_scope_valid:
        env_state = MathyEnvState(problem=text)
        reward = env.get_state_transition(env_state)
        assert text == text and is_terminal_transition(reward) == bool(is_win)



from ..poly_haystack_like_terms import PolyHaystackLikeTerms
from ...types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register

#
# Identify like terms in a haystack
#


class GymPolynomialLikeTermsHaystack(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs):
        super(GymPolynomialLikeTermsHaystack, self).__init__(
            env_class=PolyHaystackLikeTerms,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class PolynomialLikeTermsHaystackEasy(GymPolynomialLikeTermsHaystack):
    def __init__(self, **kwargs):
        super(PolynomialLikeTermsHaystackEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class PolynomialLikeTermsHaystackNormal(GymPolynomialLikeTermsHaystack):
    def __init__(self, **kwargs):
        super(PolynomialLikeTermsHaystackNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class PolynomialLikeTermsHaystackHard(GymPolynomialLikeTermsHaystack):
    def __init__(self, **kwargs):
        super(PolynomialLikeTermsHaystackHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(
    id="mathy-poly-like-terms-haystack-easy-v0",
    entry_point="mathy.envs.gym:PolynomialLikeTermsHaystackEasy",
)
safe_register(
    id="mathy-poly-like-terms-haystack-normal-v0",
    entry_point="mathy.envs.gym:PolynomialLikeTermsHaystackNormal",
)
safe_register(
    id="mathy-poly-like-terms-haystack-hard-v0",
    entry_point="mathy.envs.gym:PolynomialLikeTermsHaystackHard",
)



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""TensorFlow ops for state value learning."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
import tensorflow as tf
from . import base_ops
from . import sequence_ops


TDExtra = collections.namedtuple("td_extra", ["target", "td_error"])
TDLambdaExtra = collections.namedtuple(
    "td_lambda_extra", ["temporal_differences", "discounted_returns"]
)


def td_learning(v_tm1, r_t, pcont_t, v_t, name="TDLearning"):
    """Implements the TD(0)-learning loss as a TensorFlow op.

  The TD loss is `0.5` times the squared difference between `v_tm1` and
  the target `r_t + pcont_t * v_t`.

  See "Learning to Predict by the Methods of Temporal Differences" by Sutton.
  (https://link.springer.com/article/10.1023/A:1022633531479).

  Args:
    v_tm1: Tensor holding values at previous timestep, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    v_t: Tensor holding values at current timestep, shape `[B]`.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `v_tm1`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[v_tm1, v_t, r_t, pcont_t]], [1], name)

    # TD(0)-learning op.
    with tf.compat.v1.name_scope(name, values=[v_tm1, r_t, pcont_t, v_t]):

        # Build target.
        target = tf.stop_gradient(r_t + pcont_t * v_t)

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - v_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, TDExtra(target, td_error))


def generalized_lambda_returns(
    rewards,
    pcontinues,
    values,
    bootstrap_value,
    lambda_=1,
    name="generalized_lambda_returns",
):
    """Computes lambda-returns along a batch of (chunks of) trajectories.

  For lambda=1 these will be multistep returns looking ahead from each
  state to the end of the chunk, where bootstrap_value is used. If you pass an
  entire trajectory and zeros for bootstrap_value, this is just the Monte-Carlo
  return / TD(1) target.

  For lambda=0 these are one-step TD(0) targets.

  For inbetween values of lambda these are lambda-returns / TD(lambda) targets,
  except that traces are always cut off at the end of the chunk, since we can't
  see returns beyond then. If you pass an entire trajectory with zeros for
  bootstrap_value though, then they're plain TD(lambda) targets.

  lambda can also be a tensor of values in [0, 1], determining the mix of
  bootstrapping vs further accumulation of multistep returns at each timestep.
  This can be used to implement Retrace and other algorithms. See
  `sequence_ops.multistep_forward_view` for more info on this. Another way to
  think about the end-of-chunk cutoff is that lambda is always effectively zero
  on the timestep after the end of the chunk, since at the end of the chunk we
  rely entirely on bootstrapping and can't accumulate returns looking further
  into the future.

  The sequences in the tensors should be aligned such that an agent in a state
  with value `V` transitions into another state with value `V'`, receiving
  reward `r` and pcontinue `p`. Then `V`, `r` and `p` are all at the same index
  `i` in the corresponding tensors. `V'` is at index `i+1`, or in the
  `bootstrap_value` tensor if `i == T`.

  Subtracting `values` from these lambda-returns will yield estimates of the
  advantage function which can be used for both the policy gradient loss and
  the baseline value function loss in A3C / GAE.

  Args:
    rewards: 2-D Tensor with shape `[T, B]`.
    pcontinues: 2-D Tensor with shape `[T, B]`.
    values: 2-D Tensor containing estimates of the state values for timesteps
      0 to `T-1`. Shape `[T, B]`.
    bootstrap_value: 1-D Tensor containing an estimate of the value of the
      final state at time `T`, used for bootstrapping the target n-step
      returns. Shape `[B]`.
    lambda_: an optional scalar or 2-D Tensor with shape `[T, B]`.
    name: Customises the name_scope for this op.

  Returns:
    2-D Tensor with shape `[T, B]`
  """
    values.get_shape().assert_has_rank(2)
    rewards.get_shape().assert_has_rank(2)
    pcontinues.get_shape().assert_has_rank(2)
    bootstrap_value.get_shape().assert_has_rank(1)
    scoped_values = [rewards, pcontinues, values, bootstrap_value, lambda_]
    with tf.compat.v1.name_scope(name, values=scoped_values):
        if lambda_ == 1:
            # This is actually equivalent to the branch below, just an optimisation
            # to avoid unnecessary work in this case:
            return sequence_ops.scan_discounted_sum(
                rewards,
                pcontinues,
                initial_value=bootstrap_value,
                reverse=True,
                back_prop=False,
                name="multistep_returns",
            )
        else:
            v_tp1 = tf.concat(
                axis=0, values=[values[1:, :], tf.expand_dims(bootstrap_value, 0)]
            )
            # `back_prop=False` prevents gradients flowing into values and
            # bootstrap_value, which is what you want when using the bootstrapped
            # lambda-returns in an update as targets for values.
            return sequence_ops.multistep_forward_view(
                rewards,
                pcontinues,
                v_tp1,
                lambda_,
                back_prop=False,
                name="generalized_lambda_returns",
            )


def td_lambda(
    state_values, rewards, pcontinues, bootstrap_value, lambda_=1, name="BaselineLoss"
):
    """Constructs a TensorFlow graph computing the L2 loss for sequences.

  This loss learns the baseline for advantage actor-critic models. Gradients
  for this loss flow through each tensor in `state_values`, but no other
  input tensors. The baseline is regressed towards the n-step bootstrapped
  returns given by the reward/pcontinue sequence.

  This function is designed for batches of sequences of data. Tensors are
  assumed to be time major (i.e. the outermost dimension is time, the second
  outermost dimension is the batch dimension). We denote the sequence length
  in the shapes of the arguments with the variable `T`, the batch size with
  the variable `B`, neither of which needs to be known at construction time.
  Index `0` of the time dimension is assumed to be the start of the sequence.

  `rewards` and `pcontinues` are the sequences of data taken directly from the
  environment, possibly modulated by a discount. `state_values` are the
  sequences of (typically learnt) estimates of the values of the states
  visited along a batch of trajectories.

  The sequences in the tensors should be aligned such that an agent in a state
  with value `V` that takes an action transitions into another state
  with value `V'`, receiving reward `r` and pcontinue `p`. Then `V`, `r`
  and `p` are all at the same index `i` in the corresponding tensors. `V'` is
  at index `i+1`, or in the `bootstrap_value` tensor if `i == T`.

  See "High-dimensional continuous control using generalized advantage
  estimation" by Schulman, Moritz, Levine et al.
  (https://arxiv.org/abs/1506.02438).

  Args:
    state_values: 2-D Tensor of state-value estimates with shape `[T, B]`.
    rewards: 2-D Tensor with shape `[T, B]`.
    pcontinues: 2-D Tensor with shape `[T, B]`.
    bootstrap_value: 1-D Tensor with shape `[B]`.
    lambda_: an optional scalar or 2-D Tensor with shape `[T, B]`.
    name: Customises the name_scope for this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * temporal_differences, Tensor of shape `[T, B]`
        * discounted_returns, Tensor of shape `[T, B]`
  """
    scoped_values = [state_values, rewards, pcontinues, bootstrap_value]
    with tf.compat.v1.name_scope(name, values=scoped_values):
        discounted_returns = generalized_lambda_returns(
            rewards, pcontinues, state_values, bootstrap_value, lambda_
        )
        temporal_differences = discounted_returns - state_values
        loss = 0.5 * tf.reduce_sum(
            input_tensor=tf.square(temporal_differences), axis=0, name="l2_loss"
        )

        return base_ops.LossOutput(
            loss,
            TDLambdaExtra(
                temporal_differences=temporal_differences,
                discounted_returns=discounted_returns,
            ),
        )


def qv_max(v_tm1, r_t, pcont_t, q_t, name="QVMAX"):
    """Implements the QVMAX learning loss as a TensorFlow op.

  The QVMAX loss is `0.5` times the squared difference between `v_tm1` and
  the target `r_t + pcont_t * max q_t`, where `q_t` is separately learned
  through QV learning (c.f. `action_value_ops.qv_learning`).

  See "The QV Family Compared to Other Reinforcement Learning Algorithms" by
  Wiering and van Hasselt (2009).
  (http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.713.1931)

  Args:
    v_tm1: Tensor holding values at previous timestep, shape `[B]`.
    r_t: Tensor holding rewards, shape `[B]`.
    pcont_t: Tensor holding pcontinue values, shape `[B]`.
    q_t: Tensor of action values at current timestep, shape `[B, num_actions]`.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the batch of losses, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `target`: batch of target values for `v_tm1`, shape `[B]`.
        * `td_error`: batch of temporal difference errors, shape `[B]`.
  """
    # Rank and compatibility checks.
    base_ops.wrap_rank_shape_assert([[v_tm1, r_t, pcont_t], [q_t]], [1, 2], name)

    # The QVMAX op.
    with tf.compat.v1.name_scope(name, values=[v_tm1, r_t, pcont_t, q_t]):

        # Build target.
        target = tf.stop_gradient(
            r_t + pcont_t * tf.reduce_max(input_tensor=q_t, axis=1)
        )

        # Temporal difference error and loss.
        # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
        td_error = target - v_tm1
        loss = 0.5 * tf.square(td_error)
        return base_ops.LossOutput(loss, TDExtra(target, td_error))




import shutil
import tempfile
from unittest.mock import patch

import pytest

from ..mathy.agents.a3c import A3CAgent, A3CConfig
from ..mathy.agents.policy_value_model import (
    PolicyValueModel,
    get_or_create_policy_model,
)
from ..mathy.cli import setup_tf_env
from ..mathy.envs import PolySimplify
from ..mathy.api import Mathy
from ..mathy.models import load_model, package, load_model_from_init_py, get_model_meta
from pathlib import Path


def test_models_package_errors():
    with pytest.raises(SystemExit):
        package("model", "/fake/path", "/fake/pout")
    input_folder = Path(__file__).parent / "test_model_sm"
    output_folder = tempfile.mkdtemp()
    with pytest.raises(SystemExit):
        package("model", input_folder, output_folder, meta_path="fake")
    shutil.rmtree(output_folder)


def test_models_load_model_errors():
    with pytest.raises(ValueError):
        mt: Mathy = load_model(None)


def test_models_load_model_from_init_py_errors():
    with pytest.raises(ValueError):
        load_model_from_init_py("./fake/__init__.py")


def test_models_get_model_meta():
    with pytest.raises(ValueError):
        get_model_meta("./fake/")


def test_models_from_package():
    setup_tf_env()
    mt: Mathy = load_model("mathy_alpha_sm")
    assert mt is not None
    assert mt.model is not None
    assert mt.config is not None


def test_models_from_path():
    setup_tf_env()
    input_folder = Path(__file__).parent / "test_model_sm"
    mt: Mathy = load_model(input_folder)
    assert mt is not None
    assert mt.model is not None
    assert mt.config is not None


def test_models_package():
    setup_tf_env()
    input_folder = Path(__file__).parent / "test_model_sm"
    output_folder = tempfile.mkdtemp()
    out_dir = package(
        model_name="zote_the_mighty", input_dir=input_folder, output_dir=output_folder,
    )
    mt: Mathy = load_model(out_dir)
    assert mt is not None
    assert mt.model is not None
    assert mt.config is not None
    shutil.rmtree(output_folder)


def test_models_train_and_package():
    setup_tf_env()
    input_folder = tempfile.mkdtemp()
    output_folder = tempfile.mkdtemp()
    setup_tf_env()
    args = A3CConfig(
        max_eps=1,
        topics=["poly-combine"],
        model_dir=input_folder,
        num_workers=1,
        units=3,
        lstm_units=3,
        embedding_units=3,
    )
    agent = A3CAgent(args)
    agent.train()
    out_dir = package(
        model_name="zote_the_mighty", input_dir=input_folder, output_dir=output_folder,
    )
    mt: Mathy = load_model(out_dir)
    assert mt is not None
    assert mt.model is not None
    assert mt.config is not None
    shutil.rmtree(output_folder)
